from .do_file_section import DoFileSection
from .imported_dataset import ImportedDataset, StataVar
from .stata_utils import stata_string_escape


MAX_LABEL_LEN = 80


class LabelVariable(DoFileSection):

    DEFAULT_SETTINGS = {
        'first_paragraph_only': False,
        'remove_numbering': True,
        'stop_at_words': [],
        'stop_before_words': [],
    }

    def __init__(self, dataset: ImportedDataset, settings: dict = None,
                 populate: bool = False):
        self.label_variables = []
        super().__init__(dataset, settings, populate)

    def populate(self):
        self.label_variables.clear()
        for var in self.dataset:
            self.analyze_variable(var)

    def analyze_variable(self, var: StataVar):
        if self.should_label(var):
            self.label_variables.append(var)

    def should_label(self, var: StataVar):
        if var.is_dropped() or var.column.survey_row is None:
            return False
        if var.column.survey_row.becomes_column():
            return True
        return False

    def do_file_iter(self):
        for var in self.label_variables:
            yield self.label_variable_do(var)

    def label_variable_do(self, var: StataVar):
        varname = var.varname
        label_raw = var.column.survey_row.get_label(self.which_label,
                                                    self.extra_label)
        label_cleaned = self.clean_label(label_raw)
        if label_cleaned == '':
            return f'* LABEL SKIPPED: variable "{varname}" has no label.'
        # TODO IMPROVE THIS PART
        label_truncated = label_cleaned[:MAX_LABEL_LEN]
        label = stata_string_escape(label_truncated)
        label_variable = f'label var {varname} {label}'
        if len(label_cleaned) > 80:
            msg = (f'* LABEL TOO LONG: original label is {len(label_cleaned)} characters long. '
                   f'the last {len(label_cleaned) - 80} characters will be lost.')
            label_variable = f'{msg}\n{label_variable}'
        return label_variable

    def clean_label(self, text: str) -> str:
        new_text = text
        if self.remove_numbering:
            new_text = self.get_number_removed(text)
        return new_text

    @staticmethod
    def get_number_removed(text: str) -> str:
        split = text.split(maxsplit=1)
        if len(split) > 1 and any(ch.isdigit() for ch in split[0]):
            return split[1]
        return text

    @property
    def remove_numbering(self):
        result = self.settings['remove_numbering']
        return result

    def __repr__(self):
        """Get a representation of this object."""
        msg = f"<LabelVariable, size {len(self.label_variables)}>"
        return msg



"""Module defining OdkForm."""
import xlrd

from .choices import Choices
from .settings import Settings
from .survey import Survey
from ..error import OdkFormError


class OdkForm:
    """Define the OdkForm class.

    Instance attributes:
        path: The path to where the XlsForm is stored
        survey: The survey component
        choices: The choices component
        settings: The settings component
    """

    def __init__(self, path: str):
        """Initialize an OdkForm.

        The survey, choices, and settings are initialized separately.
        Lastly, the survey rows are associated with their choice lists.

        Args:
            path: The path to where the XlsForm is stored
        """
        self.path = path
        workbook = xlrd.open_workbook(self.path)
        self.survey = self._parse_survey(workbook)
        self.choices = self._parse_choices(workbook)
        self.settings = self._parse_settings(self.path, workbook)
        self._associate_choice_lists()

    @staticmethod
    def _parse_survey(workbook: xlrd.Book) -> Survey:
        """Parse the survey for the ODK form.

        Args:
            workbook: The xlrd book object

        Returns:
            A Survey object
        """
        try:
            sheet = workbook.sheet_by_name('survey')
            return Survey(sheet, workbook.datemode)
        except xlrd.biffh.XLRDError:
            msg = 'XlsForm file does not have required "survey" tab!'
            raise OdkFormError(msg)

    @staticmethod
    def _parse_choices(workbook: xlrd.Book) -> Choices:
        """Parse the choices for the ODK form.

        Args:
            workbook: The xlrd book object

        Returns:
            A Choices object
        """
        return Choices(workbook)

    @staticmethod
    def _parse_settings(path: str, workbook: xlrd.Book) -> Settings:
        """Parse the settings for the ODK form.

        Args:
            path: The path to where the XlsForm is stored
            workbook: The xlrd book object

        Returns:
            A Settings object
        """
        sheet = None
        try:
            sheet = workbook.sheet_by_name('settings')
        except xlrd.biffh.XLRDError:
            pass
        return Settings(path, sheet, workbook.datemode)

    def _associate_choice_lists(self) -> None:
        """Associate survey rows with their choice lists."""
        for row in self.survey:
            if row.is_select_type():
                list_name = row.get_select_list_name()
                # Future: how we get the list may change
                if row.is_select_external():
                    choice_list = self.choices.external_choices[list_name]
                else:
                    choice_list = self.choices.choices[list_name]
                row.choice_list = choice_list

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'OdkForm("{self.path}")'
        return msg



import configparser
import io

from ..destring import Destring
from ..metadata import Metadata
from ..do_file_section import DoFileSection
from ..drop_column import DropColumn
from ..encode_select_one import EncodeSelectOne
from ..label_variable import LabelVariable
from ..rename import Rename
from ..split_select_multiple import SplitSelectMultiple
from ...dataset.utils import DatasetSource


def stringify_value(value):
    if isinstance(value, DatasetSource):
        result = dataset_source_to_str(value)
    elif isinstance(value, list):
        result = '\n'.join(value)
    else:
        result = str(value)
    return result


def stringify_values(settings):
    result = {k: stringify_value(v) for k, v in settings.items()}
    return result


def str_to_list(value: str) -> list:
    if not value:
        return []
    split = value.split('\n')
    return split


def list_to_str(value: list) -> str:
    if not value:
        return ''
    result = '\n'.join(value)
    return result


def str_to_dataset_source(value: str) -> DatasetSource:
    result = DatasetSource.from_string(value)
    return result


def dataset_source_to_str(value: DatasetSource) -> str:
    result = value.name.lower()
    return result


STR_BOOL_TRUE = ['true', 'yes', 't', 'y', '1']


def destringify_value(value: str, pattern):
    if isinstance(pattern, DatasetSource):
        result = DatasetSource.from_string(value)
    elif isinstance(pattern, list):
        result = str_to_list(value)
    elif isinstance(pattern, bool):
        result = value.lower() in STR_BOOL_TRUE
    else:
        result = value
    return result


def destringify_values(settings: dict, pattern: dict, default: dict) -> dict:
    result = {}
    combined_pattern = {}
    combined_pattern.update(default)
    combined_pattern.update(pattern)
    for k, v in settings.items():
        pattern_value = combined_pattern.get(k)
        if pattern_value is not None:
            value = destringify_value(v, pattern_value)
            result[k] = value
        else:
            # Setting will not be used
            result[k] = v
    return result


class SettingsManager:
    # Knows about each of the sections
    # Can take an ini file and make a dictionary for each section
    # Can make output in ini format or JSON format
    # Can retrieve other settings in this folder

    def __init__(self, path=None):
        self.default = {}
        self.destring = {}
        self.drop_column = {}
        self.encode_select_one = {}
        self.label_variable = {}
        self.metadata = {}
        self.rename = {}
        self.split_select_multiple = {}
        if path:
            self.import_from_file(path)
        else:
            self.set_all_to_default()

    def set_all_to_empty(self):
        self.default = {}
        self.destring = {}
        self.drop_column = {}
        self.encode_select_one = {}
        self.label_variable = {}
        self.metadata = {}
        self.rename = {}
        self.split_select_multiple = {}

    def set_all_to_default(self):
        self.default = DoFileSection.BASE_DEFAULT_SETTINGS.copy()
        self.destring = Destring.DEFAULT_SETTINGS.copy()
        self.drop_column = DropColumn.DEFAULT_SETTINGS.copy()
        self.encode_select_one = EncodeSelectOne.DEFAULT_SETTINGS.copy()
        self.label_variable = LabelVariable.DEFAULT_SETTINGS.copy()
        self.metadata = Metadata.DEFAULT_SETTINGS.copy()
        self.rename = Rename.DEFAULT_SETTINGS.copy()
        self.split_select_multiple = SplitSelectMultiple.DEFAULT_SETTINGS.copy()

    def import_from_file(self, path):
        config = configparser.ConfigParser(interpolation=None)
        config.read(path)
        self.destring = destringify_values(
            config['destring'],
            Destring.DEFAULT_SETTINGS,
            DoFileSection.BASE_DEFAULT_SETTINGS
        )
        self.drop_column = destringify_values(
            config['drop_column'],
            DropColumn.DEFAULT_SETTINGS,
            DoFileSection.BASE_DEFAULT_SETTINGS
        )
        self.encode_select_one = destringify_values(
            config['encode_select_one'],
            EncodeSelectOne.DEFAULT_SETTINGS,
            DoFileSection.BASE_DEFAULT_SETTINGS
        )
        self.label_variable = destringify_values(
            config['label_variable'],
            LabelVariable.DEFAULT_SETTINGS,
            DoFileSection.BASE_DEFAULT_SETTINGS
        )
        self.metadata = destringify_values(
            config['metadata'],
            Metadata.DEFAULT_SETTINGS,
            DoFileSection.BASE_DEFAULT_SETTINGS
        )
        self.rename = destringify_values(
            config['rename'],
            Rename.DEFAULT_SETTINGS,
            DoFileSection.BASE_DEFAULT_SETTINGS
        )
        self.split_select_multiple = destringify_values(
            config['split_select_multiple'],
            SplitSelectMultiple.DEFAULT_SETTINGS,
            DoFileSection.BASE_DEFAULT_SETTINGS
        )

    def to_dict(self):
        result = {
            'destring': self.destring,
            'drop_column': self.drop_column,
            'encode_select_one': self.encode_select_one,
            'label_variable': self.label_variable,
            'metadata': self.metadata,
            'rename': self.rename,
            'split_select_multiple': self.split_select_multiple,
        }
        return result

    def get_case_preserve(self) -> bool:
        return self.metadata['case_preserve']

    def get_merge_single_repeat(self) -> bool:
        return self.metadata['merge_single_repeat']

    def get_merge_append(self) -> bool:
        return self.metadata['merge_append']

    @classmethod
    def generate_default_ini(cls, path=None):
        settings = cls()
        settings.set_all_to_default()
        config = configparser.ConfigParser(interpolation=None)
        config['DEFAULT'] = stringify_values(settings.default)
        config['destring'] = stringify_values(settings.destring)
        config['drop_column'] = stringify_values(settings.drop_column)
        config['encode_select_one'] = \
            stringify_values(settings.encode_select_one)
        config['label_variable'] = stringify_values(settings.label_variable)
        config['metadata'] = stringify_values(settings.metadata)
        config['rename'] = stringify_values(settings.rename)
        config['split_select_multiple'] = \
            stringify_values(settings.split_select_multiple)
        if path:
            with open(path, 'w') as fp:
                config.write(fp, space_around_delimiters=True)
        else:
            output = io.StringIO()
            config.write(output, space_around_delimiters=True)
            contents = output.getvalue()
            output.close()
            return contents


if __name__ == '__main__':
    # SettingsManager.generate_default_ini('/Users/jpringle/Desktop/default.ini')
    # print(SettingsManager.generate_default_ini())
    sm = SettingsManager()
    sm.import_from_file('/Users/jpringle/Desktop/default.ini')
    print(sm.to_dict())



"""A module to define the Settings class."""
import os.path

import xlrd.sheet

from .components import Worksheet


class Settings(Worksheet):
    """A class to represent the settings of an XlsForm.

    Instance attributes:
        path: The path to where the XlsForm is stored
        settings: A dictionary of settings properties and values
    """

    def __init__(self, path: str, sheet: xlrd.sheet.Sheet,
                 datemode: int = Worksheet.DEFAULT_DATEMODE):
        """Initialize the Settings object.

        If settings does not define form_title or form_id, they are
        computed using the path to the XlsForm file.

        Args:
            path: The path to where the XlsForm is stored
            sheet: The xlrd sheet that stores the settings
            datemode: The datemode for the workbook
        """
        self.path = path
        self.settings = {}
        if sheet is not None:
            self._parse_settings(sheet, datemode)
        self.settings['form_title'] = self.form_title
        self.settings['form_id'] = self.form_id

    @property
    def form_title(self):
        """Get the settings form_title."""
        form_title = self.settings.get('form_title')
        if not form_title:
            _, full_filename = os.path.split(self.path)
            filename, _ = os.path.splitext(full_filename)
            form_title = filename
        return form_title

    @property
    def form_id(self):
        """Get the settings form_id."""
        form_id = self.settings.get('form_id')
        if not form_id:
            form_id = self.form_title
        return form_id

    def _parse_settings(self, sheet: xlrd.sheet.Sheet, datemode: int):
        """Get settings.

        Pre-condition: sheet is not None.

        Args:
            sheet: The xlrd sheet that stores the settings
            datemode: The datemode for the workbook
        """
        try:
            keys = self.get_header(sheet, datemode)
            values = self.get_row_values(sheet, 1, datemode)
            self.settings = {k: v for k, v in zip(keys, values)}
        except IndexError:
            pass

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<Settings {repr(self.settings)}>'
        return msg



"""A module defining the Column class."""
from ..odkform.survey import SurveyRow


class Column:
    """Describe a column in the dataset.

    Instance attributes:
        column_name: The column name in the original CSV
        survey_row: The SurveyRow associated with this Column
        repeat_dataset: The repeat group Dataset associated with this
            column, if there is one.
    """

    def __init__(self, column_name: str, survey_row: SurveyRow):
        """Initialize a Column instance.

        Args:
            column_name: The column name in the original CSV
            survey_row: The SurveyRow associated with this Column
        """
        self.column_name = column_name
        self.survey_row = survey_row
        self.repeat_dataset = None

    def from_single_odk_row(self) -> bool:
        has_survey_row = self.survey_row is not None
        return has_survey_row and self.survey_row.becomes_single_column()

    def based_on_odk_row(self) -> bool:
        return self.survey_row is not None

    def get_survey_row(self) -> SurveyRow:
        return self.survey_row

    def get_odk_type(self):
        if self.survey_row is None:
            return None
        return self.survey_row.get_type()

    def get_odk_name(self):
        if self.survey_row is None:
            return None
        return self.survey_row.row_name

    def is_numeric(self):
        if self.survey_row is None:
            return False
        return self.survey_row.is_numeric_type()

    def __hash__(self):
        """Make a hash based on the column name and the survey row."""
        return hash((self.column_name, self.survey_row))

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'Column({self.column_name}, {self.survey_row!r})'
        return msg



"""A module defining custom exceptions for the package.

Each subpackage has a base exception class.
"""


class OdkFormError(Exception):
    """The base exception class for the odkform subpackage."""


class MismatchedGroupOrRepeatError(OdkFormError):
    """An exception for parsing group or repeats."""


class LabelNotFoundError(OdkFormError):
    """An excpetion for determining the label for an XlsFormRow."""


class DatasetError(Exception):
    """The base exception class for the dataset subpackage."""


class DoFileError(Exception):
    """The base exception class for the dofile subpackage."""


class RenameNotApplicableError(Exception):
    """An exception when trying to apply a rename that doesn't apply."""


class RenameNotSupportedError(Exception):
    """An exception when trying to apply a rename that doesn't apply."""



"""A module for the DoFileCollection class."""
from typing import List

from .do_file import DoFile
from .settings import SettingsManager
from ..dataset import DatasetCollection


class DoFileCollection:
    """A class to represent do files coming from the same ODK file.

    The DoFileCollection class corresponds to the DatasetCollection
    class. Thus, for each Dataset in a DatasetCollection, a DoFile is
    created and managed in this class.

    One small difference is that if there is a single repeat group in
    the source dataset, then there is an option to merge them together
    and have one DoFile for both.

    Instance attributes:
        dataset_collection: A reference to the input dataset collection
        settings: A reference to the input settings. If input is None,
            then this is the default settings
        do_files: A list of DoFiles, the members in this collection
    """

    def __init__(self, dataset_collection: DatasetCollection,
                 settings: SettingsManager = None):
        """Initialize a DoFileCollection.

        Args:
            dataset_collection: A dataset that is defined by an XLSForm
            settings: The settings used to control how do files are
                initialized
        """
        self.dataset_collection = dataset_collection
        self.settings = settings if settings else SettingsManager()
        self.do_files: List[DoFile] = []

        merge_in_settings = self.settings.get_merge_single_repeat()
        dataset_can_merge = self.dataset_collection.can_merge_single_repeat()
        if merge_in_settings and dataset_can_merge:
            do_file = DoFile(dataset_collection.primary, settings)
            self.do_files.append(do_file)
        else:
            for dataset in dataset_collection.get_datasets():
                do_file = DoFile(dataset, settings)
                self.do_files.append(do_file)

    @classmethod
    def from_file(cls, path: str, dataset_source: str = 'briefcase',
                  settings_path: str = None):
        """Initialize an instance based on input file paths.

        Args:
            path: The path to the source XLSForm
            dataset_source: Where the dataset source comes from
            settings_path: The path to the settings file

        Returns:
            An initialized do file collection instance.
        """
        dataset_collection = DatasetCollection.from_file(path, dataset_source)
        settings = SettingsManager(settings_path)
        return cls(dataset_collection, settings)

    def render(self) -> str:
        """Write all do files out in sequence.

        Returns:
            A string with all do files joined together
        """
        rendered = (do_file.render() for do_file in self.do_files)
        joined_do = '\n\n\n'.join(rendered)
        return joined_do

    def write_out(self, path: str):
        """Write out the result of the `render` routine.

        Args:
            path: Path to where the result should be saved
        """
        with open(path, mode='w', encoding='utf-8') as file:
            for do_file in self.do_files:
                render = do_file.render()
                file.write(render)

    def write_out_singly(self, path_to_dir: str):
        """Write out the do files in this collection to a directory.

        Args:
            path_to_dir: Path to the output directory
        """
        # TODO: Fill this in. Filename should be same as CSV but end with .do?



from enum import Enum
import re
from typing import List

from .do_file_section import DoFileSection
from .imported_dataset import ImportedDataset, StataVar
from .stata_utils import VARNAME_CHARACTERS, is_valid_stata_varname
from .templates import env
from ..error import RenameNotApplicableError, RenameNotSupportedError


RENAME_ODK_NAME_SECT = env.get_template('rename_odk_name_sect.do')


class Rename(DoFileSection):

    DEFAULT_SETTINGS = {
        'direct_rename': [],
        'rename_to_odk_name': True
    }

    def __init__(self, dataset: ImportedDataset, settings: dict = None,
                 populate: bool = True):
        self.direct_rules: List[RenameRule] = []
        self.odk_name_rules: List[RenameRule] = []
        super().__init__(dataset, settings, populate)

    def on_settings_updated(self):
        rules = self.parse_direct_rename()
        self.direct_rules.extend(rules)

    def parse_direct_rename(self):
        rules = []
        for line in self.direct_rename:
            split = line.split(maxsplit=1)
            rule = RenameRule(split[0], split[1])
            rules.append(rule)
        return rules

    def populate(self):
        self.dataset.set_all_varnames_to_original()
        self.odk_name_rules.clear()
        for var in self.dataset:
            self.analyze_variable(var)

    def analyze_variable(self, var: StataVar):
        # Analyzing a variable, we learn if
        # 1) we need to make a RenameRule back to original odk name
        # 2) we need to update the varname of the StataVar (based on direct rules)
        if var.is_dropped() or var.get_odk_name() is None:
            return
        if self.rename_to_odk_name:
            from_single_odk_row = var.from_single_odk_row()
            imported_name = var.orig_varname
            # TODO: WORK WITH STATA CLEANED VARNAME
            cleaned_stata_odk_name = var.get_odk_name()[:32]
            should_change = imported_name != cleaned_stata_odk_name
            if from_single_odk_row and should_change:
                # If renaming to ODK name, then make direct rename rule
                rule = RenameRule(var.orig_varname, var.get_odk_name())
                self.odk_name_rules.append(rule)
                var.set_current_varname(var.get_odk_name())
        new_name = self.apply_rules(var.varname)
        var.set_current_varname(new_name)

    def do_file_iter(self):
        if self.rename_to_odk_name:
            yield
        yield ''

    def apply_rules(self, varname: str) -> str:
        """Apply rename rules to input varname.

        Args:
            varname: A Stata varname

        Returns:
            The varname that would be found in the dataset after
            applying all rename rules.
        """
        current = varname
        for rule in self.direct_rules:
            try:
                current = rule.apply(current)
            except RenameNotApplicableError:
                # Current varname after this rule remains the same
                pass
        return current

    def get_varname(self, varname: str) -> str:
        """Apply rename rules to input varname.

        Args:
            varname: A STATA varname

        Returns:
            The varname that would be found in the dataset after
            applying the renaming in this object.
        """
        # TODO: Fix this. It is a stub
        return varname

    def rename_to_odk_name_do(self) -> str:
        result = RENAME_ODK_NAME_SECT.render(
            odk_name_rules=self.odk_name_rules
        )
        return result

    @property
    def direct_rename(self):
        result = self.settings['direct_rename']
        return result

    @property
    def rename_to_odk_name(self):
        result = self.settings['rename_to_odk_name']
        return result

    def __repr__(self):
        size = len(self.direct_rules) + len(self.odk_name_rules)
        return f'<Rename, size: {size}>'


class RenameRule:
    # A rule has the old and the new parts
    # The old and new parts get parsed

    SPECIAL_CHARS = '*?#.='

    def __init__(self, old: str, new: str):
        self.old = old
        self.old_parsed = RenameParser(self.old)
        self.new = new
        self.new_parsed = RenameParser(self.new)

    def affects(self, varname: str) -> bool:
        match = re.match(self.old_parsed.regex, varname)
        return bool(match)

    def apply(self, varname: str) -> str:
        if not self.affects(varname):
            raise RenameNotApplicableError(repr(self), varname)
        match = re.match(self.old_parsed.regex, varname)
        matches = list(zip(match.groups(), self.old_parsed))
        new_chunks = []
        for token in self.new_parsed:
            if token.type == RenameType.DIRECT:
                new_chunks.append(token.string())
            elif token.type == RenameType.ASTERISK:
                found = -1
                for i, item in enumerate(matches):
                    found = i
                    matched, token = item
                    if token.type == RenameType.ASTERISK:
                        result = matched
                        new_chunks.append(result)
                        break
                if found >= 0:
                    matches.pop(found)
        return ''.join(new_chunks)

    def direct_rename(self, varname: str) -> str:
        old_is_varname = self.is_stata_varname(self.old)
        new_is_varname = self.is_stata_varname(self.new)
        if varname == self.old and old_is_varname and new_is_varname:
            return self.new
        else:
            raise RenameNotApplicableError(repr(self), varname)

    @staticmethod
    def has_special_char(name: str) -> bool:
        has = any(char in name for char in RenameRule.SPECIAL_CHARS)
        return has

    @staticmethod
    def is_stata_varname(name: str) -> bool:
        return is_valid_stata_varname(name)

    def __repr__(self):
        return f'Rename("{self.old}", "{self.new}")'


class RenameParser:

    ASTERISK_REGEX = '([a-zA-Z0-9_]*)'

    def __init__(self, name: str):
        # Make tokens out of everything that matches
        # Match those tokens to the result (both should be the same length)
        self.name = name
        self.tokens = []
        current_token = RenameToken()
        for char in name:
            if current_token.is_compatible(char):
                current_token.add(char)
            else:
                self.tokens.append(current_token)
                current_token = RenameToken(char)
        self.tokens.append(current_token)
        self.regex = self.generate_regex()

    def generate_regex(self):
        regex_chunks = []
        for token in self.tokens:
            if token.type == RenameType.DIRECT:
                direct_match = f'({token.string()})'
                regex_chunks.append(direct_match)
            elif token.type == RenameType.ASTERISK:
                regex_chunks.append(self.ASTERISK_REGEX)
            else:
                raise NotImplementedError()
        return ''.join(regex_chunks)

    def __iter__(self):
        return iter(self.tokens)

    def __len__(self):
        return len(self.tokens)

    def __repr__(self):
        return f'<RenameParser("{self.name}"), size: {len(self.tokens)}>'


class RenameToken:

    def __init__(self, first_char: str = None):
        self.chars = []
        self.type = None
        if first_char:
            self.chars.append(first_char)
            self.set_type(first_char)

    def string(self):
        return ''.join(self.chars)

    def is_empty(self):
        return len(self.chars) == 0

    def type_not_set(self):
        return self.type is None

    def is_compatible(self, char):
        if self.is_empty():
            return True
        elif self.type == RenameType.DIRECT:
            return char in VARNAME_CHARACTERS
        elif self.type == RenameType.ASTERISK:
            return False
        else:
            return NotImplementedError(char)

    def add(self, char):
        self.chars.append(char)
        if self.type_not_set():
            self.set_type(char)

    def set_type(self, char):
        if char == '*':
            self.type = RenameType.ASTERISK
        # elif char == '.':
        #     self.type = RenameType.PERIOD
        # elif char == '=':
        #     self.type = RenameType.EQUALS
        # elif char == '#':
        #     self.type = RenameType.OCTOTHORPE
        elif char in VARNAME_CHARACTERS:
            self.type = RenameType.DIRECT
        else:
            raise RenameNotSupportedError(char)

    def __repr__(self):
        return f'<RenameToken, {self.type}, "{self.string()}">'


class RenameType(Enum):
    """An enumeration of possible rename match types."""
    DIRECT = 1
    ASTERISK = 2
    PERIOD = 3
    EQUALS = 4
    OCTOTHORPE = 5



"""A collection of useful Stata-related functions.

Module attributes:
    STATA_VARNAME_REGEX: A regular expression for Stata varnames
    STATA_VARNAME_REGEX_PROG: A compiled regular expression for Stata
        varnames based on STATA_VARNAME_REGEX
    VARNAME_CHARACTERS: The full set of valid Stata varname characters
    LabelDefineOption: A named tuple with number and label attributes.
        This is used for encoding select_one variables.
"""
import re
import string


def clean_stata_varname(column_name: str) -> str:
    """Clean a Stata varname.

    Strip out bad characters first, then truncate to 32 characters.

    Args:
        column_name: An uncleaned column name

    Returns:
        A cleaned Stata varname
    """
    cleaned = varname_strip(column_name)
    letter_start = varname_valid_start(cleaned)
    truncated = varname_truncate(letter_start)
    return truncated


VARNAME_CHARACTERS = set(string.ascii_letters + string.digits + '_')


def varname_strip(varname: str) -> str:
    """Strip out invalid characters from a Stata varname."""
    new_varname = ''.join(i for i in varname if i in VARNAME_CHARACTERS)
    return new_varname


VARNAME_START_CHARACTERS = set(string.ascii_letters + '_')


def varname_valid_start(varname: str) -> str:
    """Ensure a Stata varname starts with a letter.

    This routine checks the first character of a supplied varname, and
    if needed, it prepends a 'v' at the beginning and returns.

    A valid first character is in the class [_a-zA-Z].
    """
    if varname == '':
        return 'v'
    new_varname = varname
    first_char = new_varname[0]
    if first_char not in VARNAME_START_CHARACTERS:
        new_varname = f'v{new_varname}'
    return new_varname


VARNAME_MAX_LEN = 32
LABEL_MAX_LEN = 80


def varname_truncate(varname: str) -> str:
    """Truncate a Stata varname down to 32 characters, if needed."""
    if len(varname) > VARNAME_MAX_LEN:
        return varname[:VARNAME_MAX_LEN]
    return varname


ALPHANUM_CHARACTERS = set(string.ascii_letters + string.digits)


def strip_non_alphanum(varname: str) -> str:
    """Strip out non-alphanumeric characters from a Stata varname."""
    new_varname = ''.join(i for i in varname if i in ALPHANUM_CHARACTERS)
    return new_varname


def gen_anonymous_varname(column_number: int) -> str:
    """Generate a Stata varname based on the column number.

    Stata columns are 1-indexed.
    """
    return f'v{column_number}'


def gen_tmp_varname(varname: str) -> str:
    """Generate a temporary Stata varname based on a given varname."""
    stem = varname[:30]
    return f'{stem}V2'


STATA_VARNAME_REGEX = '[_a-zA-Z][_a-zA-Z0-9]{,31}'
STATA_VARNAME_REGEX_PROG = re.compile(STATA_VARNAME_REGEX)


def is_valid_stata_varname(varname: str) -> bool:
    """Return True if and only if varname is valid in Stata.

    Currently, this function checks against the basic regular
    expression. In the future, it will check against a list of
    prohibited Stata varnames, according to Stata documentation.
    """
    valid = bool(STATA_VARNAME_REGEX_PROG.fullmatch(varname))
    return valid


def safe_stata_string_quote(text: str) -> str:
    """Safely quote a string for Stata.

    Use this when the string should not be changed.

    Args:
        text: A string for a Stata argument

    Returns:
        A quoted string. If there is a quotation mark, then we make use
        of Stata compound quotation marks.
    """
    if '"' in text:
        # Compound double quotes
        return f"""`"{text}"'"""
    return f'"{text}"'


def stata_string_escape(text: str) -> str:
    """Escape a string for Stata.

    Use this when a string needs to be escaped for a single line.

    Args:
        text: A string for a Stata argument

    Returns:
        A quoted string.
    """
    new_text = text
    new_text = new_text.replace('\n', ' ')
    new_text = new_text.replace('\t', ' ')
    new_text = new_text.replace('"', "'")
    return f'"{new_text}"'


def make_invalid_varname_comment(varname: str):
    """Make a comment about a Stata varname being invalid."""
    return f'* Invalid STATA varname: {varname}'


def get_varname_comments(*args):
    """Make invalid Stata varname comments for many varnames."""
    comments = [make_invalid_varname_comment(i) for i in args if not
                is_valid_stata_varname(i)]
    return comments



"""Describe ODK form components that can serve as base classes.

Module attributes:
    Worksheet: A class to describe a generic worksheet
    XlsFormRow: A class to describe a row in the XlsForm
"""
import datetime
from typing import Tuple

import xlrd

from ..error import LabelNotFoundError


class Worksheet:
    """A class to represent a single tab in an XlsForm.

    The Worksheet class is a base class for other ODK form building
    blocks.

    Class attributes:
        DEFAULT_DATEMODE: the default date mode for xlrd.
    """

    DEFAULT_DATEMODE=1

    @staticmethod
    def get_header(sheet: xlrd.sheet.Sheet,
                   datemode: int = DEFAULT_DATEMODE) -> Tuple[str]:
        """Get the header row for a sheet.

        Returns:
            A tuple of str for each cell in the header row
        """
        values = Worksheet.get_row_values(sheet, 0, datemode)
        return tuple(str(val) for val in values)

    @staticmethod
    def get_row_values(sheet: xlrd.sheet.Sheet, rowx: int,
                       datemode: int = DEFAULT_DATEMODE) -> tuple:
        """Get the values from a row in a sheet.

        Returns:
            A tuple of cell values. The cell values can be any of str,
            bool, int, or float.
        """
        row = sheet.row(rowx)
        values = (Worksheet.cell_to_value(cell, datemode) for cell in row)
        return tuple(values)

    @staticmethod
    def cell_to_value(cell: xlrd.sheet.Cell, datemode: int):
        """Get the value of a single cell.

        Returns:
            If the cell is a date or an error, that is converted to
            string and returned. Otherwise, a cell can be one of str,
            bool, int, or float.
        """
        if cell.ctype in (xlrd.XL_CELL_EMPTY, xlrd.XL_CELL_BLANK):
            return ''
        elif cell.ctype == xlrd.XL_CELL_TEXT:
            return cell.value.strip()
        elif cell.ctype == xlrd.XL_CELL_BOOLEAN:
            return bool(cell.value)
        elif cell.ctype == xlrd.XL_CELL_NUMBER:
            int_value = int(cell.value)
            if int_value == cell.value:
                return int_value
            return cell.value
        elif cell.ctype == xlrd.XL_CELL_DATE:
            datetime_or_time_only = xlrd.xldate_as_tuple(cell.value, datemode)
            if datetime_or_time_only[:3] == (0, 0, 0):
                return str(datetime.time(*datetime_or_time_only[3:]))
            return str(datetime.datetime(*datetime_or_time_only))
        elif cell.ctype == xlrd.XL_CELL_ERROR:
            return f'#ERROR({cell.value})'
        else:
            return str(cell.value)


class XlsFormRow:
    """A class to represent a row in the survey or choices tab.

    XlsFormRow is a base class for other ODK form building blocks. The
    attributes here are meant to be read-only.

    Instance attributes:
        rowx: The 0-indexed row in the sheet
        row_name: The value for the name column of this row
        row_header: The headers
        row_values: The values of each cell in this row
        row_dict: An easy lookup to get the value under a header
    """

    def __init__(self, rowx: int, row_name: str, row_header: Tuple[str],
                 row_values: list, row_dict: dict):
        """Initialize an XlsFormRow.

        Args:
            rowx: The 0-indexed row in the sheet
            row_name: The value for the name column of this row
            row_header: The headers
            row_values: The values of each cell in this row
            row_dict: An easy lookup to get the value under a header
        """
        self.rowx = rowx
        self.row_name = row_name
        self.row_header = row_header
        self.row_values = row_values
        self.row_dict = row_dict

    def get_label(self, which_label: str, extra_label: str) -> str:
        """Get the label for this row.

        Args:
            which_label: Which column the label should be by default.
                This parameter can take the special value of
                "first_label" to indicate the first column that starts
                with "label".
            extra_label: A column to check first. If there is something
                there, it is used. Otherwise, the search is performed
                according to `which_label`.

        Returns:
            The label that was discovered.

        Raises:
            LabelNotFoundError: If no column is found in which to find
                a label.
        """
        result = self.row_dict.get(extra_label)
        if result:
            # We will use this label
            pass
        elif which_label == 'first_label':
            header = self.row_header
            first_label = next((i for i in header if i.startswith('label')),
                               None)
            if first_label is None:
                raise LabelNotFoundError()
            result = self.row_dict[first_label]
        elif which_label in self.row_dict:
            result = self.row_dict[which_label]
        else:
            raise LabelNotFoundError()
        str_result = str(result)
        return str_result

    def __hash__(self):
        """Make a hash based on row number and row name."""
        return hash((self.rowx, self.row_name))

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<XlsFormRow "{self.row_name}" at row {self.rowx + 1}>'
        return msg



from abc import ABC, abstractmethod

from .imported_dataset import ImportedDataset
from .imported_dataset import StataVar
from ..dataset.utils import DatasetSource


class DoFileSection(ABC):

    DEFAULT_DATASET_SOURCE = DatasetSource.BRIEFCASE
    DEFAULT_FIRST_LABEL = 'first_label'
    DEFAULT_EXTRA_LABEL = 'o2s_label'

    BASE_DEFAULT_SETTINGS = {
        'skip': False,
        'omit': False,
        'dataset_source': DEFAULT_DATASET_SOURCE,
        'which_label': DEFAULT_FIRST_LABEL,
        'extra_label': DEFAULT_EXTRA_LABEL,
    }

    @property
    @abstractmethod
    def DEFAULT_SETTINGS(self):
        return {}

    def __init__(self, dataset: ImportedDataset, settings: dict = None,
                 populate: bool = False):
        self.dataset = dataset
        self.settings = dict(self.BASE_DEFAULT_SETTINGS)
        self.settings.update(self.DEFAULT_SETTINGS)
        if settings:
            self.settings.update(settings)
        self.on_settings_updated()
        if populate:
            self.populate()

    @abstractmethod
    def populate(self):
        pass

    @abstractmethod
    def analyze_variable(self, var: StataVar):
        pass

    @abstractmethod
    def do_file_iter(self):
        pass

    def on_settings_updated(self):
        pass

    def initialize_settings(self, settings: dict, populate: bool = False):
        if settings:
            self.settings = dict(self.BASE_DEFAULT_SETTINGS)
            self.settings.update(self.DEFAULT_SETTINGS)
            self.settings.update(settings)
            if populate:
                self.populate()

    def update_settings(self, settings: dict, populate: bool = False):
        if settings:
            self.settings.update(settings)
            new_settings = dict(self.settings)
            self.initialize_settings(new_settings, populate)

    @property
    def skip(self):
        return self.settings['skip']

    @property
    def omit(self):
        return self.settings['omit']

    @property
    def dataset_source(self):
        result = self.settings['dataset_source']
        return result

    @property
    def which_label(self):
        """From which column to get labels to use by default."""
        return self.settings['which_label']

    @property
    def extra_label(self):
        """From which column to get labels if they exist."""
        return self.settings['extra_label']






"""Module for the DoFile class."""
from .templates import env
from .destring import Destring
from .drop_column import DropColumn
from .encode_select_one import EncodeSelectOne
from .imported_dataset import ImportedDataset
from .label_variable import LabelVariable
from .metadata import Metadata
from .rename import Rename
from .settings import SettingsManager
from .split_select_multiple import SplitSelectMultiple
from ..dataset.dataset import Dataset


class DoFile:
    """A class to represent a Stata do file.

    A do file is determined by the source dataset and the settings that
    are applied to generate the do file.

    Instance attributes:
        settings: The settings for creating this do file. If input
            settings is None, then default settings are used
        original_dataset: A reference to the input dataset
        dataset: The imported dataset for Stata
        metadata: Metadata about the dataset and import
        drop_column: A do file section for dropping variables
        rename: A do file section for renaming variables
        destring: A do file section for destring-ing variables
        encode_select_one: A do file section for encoding select_one
            variables
        split_select_multiple: A do file section for splititng
            select_multiple variables
        label_variable: A do file section for labeling variables
    """

    def __init__(self, dataset: Dataset,
                 settings: SettingsManager = None):
        """Initialize a Stata do file.

        Args:
            dataset: A dataset representing source CSV
            settings: The settings for creating this do file
        """
        self.settings = settings if settings is not None else SettingsManager()
        self.original_dataset = dataset
        self.dataset = self.get_imported_dataset()

        metadata_settings = self.settings.metadata
        self.metadata = Metadata(self.dataset, metadata_settings)
        drop_column_settings = self.settings.drop_column
        self.drop_column = DropColumn(self.dataset, drop_column_settings,
                                      populate=True)
        rename_settings = self.settings.rename
        self.rename = Rename(self.dataset, rename_settings, populate=True)
        destring_settings = self.settings.destring
        self.destring = Destring(self.dataset, destring_settings, populate=True)
        encode_select_one_settings = self.settings.encode_select_one
        self.encode_select_one = EncodeSelectOne(
            self.dataset, encode_select_one_settings, populate=True
        )
        split_select_multiple_settings = self.settings.split_select_multiple
        self.split_select_multiple = SplitSelectMultiple(
            self.dataset, split_select_multiple_settings, populate=True
        )
        label_variable_settings = self.settings.label_variable
        self.label_variable = LabelVariable(
            self.dataset, label_variable_settings, populate=True
        )

    def get_imported_dataset(self) -> ImportedDataset:
        """Return an ImportedDataset instance based on the source data."""
        case_preserve = self.settings.get_case_preserve()
        merge_single_repeat = self.settings.get_merge_single_repeat()
        merge_append = self.settings.get_merge_append()
        if self.original_dataset.is_repeat_dataset():
            merge_single_repeat = False
        dataset = ImportedDataset(self.original_dataset, case_preserve,
                                  merge_single_repeat, merge_append)
        return dataset

    def render(self) -> str:
        """Write this do file.

        Returns:
            The entire do-file as a string
        """
        template = env.get_template('base.do')
        result = template.render(
            metadata=self.metadata,
            drop_column=self.drop_column,
            rename=self.rename,
            destring=self.destring,
            encode_select_one=self.encode_select_one,
            split_select_multiple=self.split_select_multiple,
            ssm_details=self.split_select_multiple.get_ssm_details(),
            label_variable=self.label_variable,
        )
        return result

    def write_out(self, path: str):
        """Write out the result of the `render` routine.

        Args:
            path: Path to where the result should be saved
        """
        render = self.render()
        with open(path, mode='w', encoding='utf-8') as file:
            file.write(render)



from typing import List

from .do_file_section import DoFileSection
from .stata_utils import is_valid_stata_varname, make_invalid_varname_comment
from .imported_dataset import ImportedDataset, StataVar


class DropColumn(DoFileSection):

    DEFAULT_SETTINGS = {
        'types_to_drop': ['note'],
        'odk_names_to_drop': [],
        'odk_names_not_to_drop': [],
    }

    def __init__(self, dataset: ImportedDataset, settings: dict = None,
                 populate: bool = True):
        self.drop: List[StataVar] = []
        super().__init__(dataset, settings, populate)

    def populate(self):
        # TODO: Do we need a drop list and to keep track in the ImportedDataset?
        self.dataset.keep_all()
        self.drop.clear()
        for var in self.dataset:
            self.analyze_variable(var)

    def analyze_variable(self, var: StataVar):
        should_drop = False
        odk_type = var.get_odk_type()
        if self.is_drop_type(odk_type):
            should_drop = True
        if self.is_dropped_odk_name(var.get_odk_name()):
            should_drop = True
        if self.is_kept_odk_name(var.get_odk_name()):
            should_drop = False
        if should_drop:
            var.drop()
            self.drop.append(var)

    def dropped_vars_iter(self):
        return (var.orig_varname for var in self.drop)

    def do_file_iter(self):
        for var in self.drop:
            varname = var.orig
            if not is_valid_stata_varname(varname):
                comment = make_invalid_varname_comment(varname)
                yield comment
            yield f'drop {varname}'

    def is_drop_type(self, row_type: str) -> bool:
        return row_type in self.types_to_drop

    def is_dropped_odk_name(self, name: str) -> bool:
        return name in self.odk_names_to_drop

    def is_kept_odk_name(self, name: str) -> bool:
        return name in self.odk_names_not_to_drop

    @property
    def types_to_drop(self):
        return self.settings['types_to_drop']

    @property
    def odk_names_to_drop(self):
        return self.settings['odk_names_to_drop']

    @property
    def odk_names_not_to_drop(self):
        return self.settings['odk_names_not_to_drop']

    def __repr__(self):
        msg = f'<DropColumn, size {len(self.drop)}>'
        return msg



class VarnameManager:

    def __init__(self):
        pass

    def get_temp_v2_varname(self, varname: str) -> str:
        """Get a temporary v2 varname.

        This is used when generating a new varname that is used for only a
        little while.

        Args:
            varname: A Stata varname

        Returns:
            The varname to be used as a temporary varname
        """
        varname_stem = varname[:30]
        new_varname = f'{varname_stem}V2'
        return new_varname

    def get_label_varname(self, choice_list_name: str) -> str:
        """Take input choice list name and return Stata varname.

        A choice list name might not be a valid Stata identifier. This
        method returns a valid Stata identifier.

        Args:
            choice_list_name: The name of the choice list

        Returns:
            A Stata varname that is used for a "label define" command
        """
        new_varname = choice_list_name[:32]
        return new_varname



from collections import namedtuple
from dataclasses import dataclass
from typing import List

from .do_file_section import DoFileSection
from .imported_dataset import ImportedDataset, StataVar
from .stata_utils import (clean_stata_varname, gen_tmp_varname,
                          stata_string_escape, strip_non_alphanum,
                          varname_strip, VARNAME_MAX_LEN, LABEL_MAX_LEN)
from .templates import env
from ..odkform.choices import NumberNameChoice


@dataclass(frozen=True)
class SplitSettings:
    """General settings related to individual splits."""
    choices_to_exclude: str
    which_label: str
    extra_label: str
    strict_numbering: bool
    number_column: str


@dataclass(frozen=True)
class GenBinary:
    """Data needed to define a generated binary variable."""
    original_varname: str
    padded_varname: str
    choice_name: str
    binary_varname: str
    binary_label: str


@dataclass(frozen=True)
class NumberNameLabel:
    """Relevant data about choice options."""
    number: int
    name: str
    label: str


SPLIT_SELECT_MULTIPLE_UNIT = env.get_template('split_select_multiple_unit.do')


class SsmUnit:

    def __init__(self, select_multiple: StataVar, split_method: str,
                 binary_label: str, split_settings: SplitSettings):
        self.select_multiple = select_multiple
        self.split_method = split_method
        self.binary_label = binary_label
        self.split_settings = split_settings

        self.original_varname = self.select_multiple.varname
        self.padded_varname = gen_tmp_varname(self.original_varname)
        self.select_multiple_label = self.get_select_multiple_label()
        self.choices = self.get_choices()
        self.gen_binaries = self.get_gen_binaries()

    def get_select_multiple_label(self) -> str:
        """Get the desired label for the select_multiple"""
        survey_row = self.select_multiple.get_survey_row()
        prompt_label = survey_row.get_label(self.split_settings.which_label,
                                            self.split_settings.extra_label)
        return prompt_label

    def get_choices(self) -> List[NumberNameLabel]:
        """Get filtered choices based on choices_to_exclude."""
        #import pdb; pdb.set_trace()
        exclude = self.split_settings.choices_to_exclude
        exclude_numbers = exclude == SplitSelectMultiple.CHOICE_EXCLUDE_NUMBERS
        exclude_negative_numbers = exclude == \
            SplitSelectMultiple.CHOICE_EXCLUDE_NEGATIVE_NUMBERS
        all_choices = self.get_all_choices()
        result = []
        for number, name, choice in all_choices:
            try:
                int_name = int(name)
                if exclude_numbers:
                    continue
                elif exclude_negative_numbers and int_name < 0:
                    continue
            except ValueError:
                pass
            label = choice.get_label(self.split_settings.which_label,
                                     self.split_settings.extra_label)
            result.append(NumberNameLabel(number, name, label))
        return result

    def get_all_choices(self) -> List[NumberNameChoice]:
        """Get all choices and their numbers and names."""
        choice_list = self.select_multiple.get_survey_row().choice_list
        number_column = self.split_settings.number_column
        if self.split_settings.strict_numbering:
            return choice_list.get_choices_strictly_numbered(number_column)
        else:
            return choice_list.get_choices_flexibly_numbered(number_column)

    def get_gen_binaries(self):
        """Get the GenBinary objects for this unit."""
        gen_binaries = []
        names = (choice.name for choice in self.choices)
        binary_varnames = self.gen_varnames()
        binary_labels = self.gen_labels()
        zipped = zip(names, binary_varnames, binary_labels)
        for choice_name, binary_varname, binary_label in zipped:
            gen_binary = GenBinary(self.original_varname, self.padded_varname,
                                   choice_name, binary_varname, binary_label)
            gen_binaries.append(gen_binary)
        return gen_binaries

    def gen_varnames(self) -> List[str]:
        split_method = self.split_method
        if split_method == SplitSelectMultiple.METHOD_APPEND_NAME:
            return self.gen_varnames_append_name()
        elif split_method == SplitSelectMultiple.METHOD_APPEND_NUMBER:
            return self.gen_varnames_append_number()
        elif split_method == SplitSelectMultiple.METHOD_NAME_ONLY:
            return self.gen_varnames_name_only()
        else:
            raise ValueError(split_method)

    def gen_varnames_append_name(self) -> List[str]:
        suffixes = []
        for choice in self.choices:
            name = choice.name
            uncut_suffix = strip_non_alphanum(name)
            suffix = uncut_suffix[:(VARNAME_MAX_LEN - 2)]
            suffixes.append(suffix)
        max_len = max(len(suffix) for suffix in suffixes)
        uncut_stem = clean_stata_varname(self.original_varname)
        stem = uncut_stem[:(VARNAME_MAX_LEN - max_len - 1)]
        varnames = [f'{stem}_{suffix}' for suffix in suffixes]
        return varnames

    def gen_varnames_append_number(self) -> List[str]:
        suffixes = []
        for choice in self.choices:
            number = choice.number
            suffix = str(abs(number))
            suffixes.append(suffix)
        max_len = max(len(suffix) for suffix in suffixes)
        uncut_stem = clean_stata_varname(self.original_varname)
        stem = uncut_stem[:(VARNAME_MAX_LEN - max_len - 1)]
        varnames = [f'{stem}_{suffix}' for suffix in suffixes]
        return varnames

    def gen_varnames_name_only(self) -> List[str]:
        return [clean_stata_varname(item.name) for item in self.choices]

    def gen_labels(self) -> List[str]:
        gen_labels = []
        uncut_stem = self.select_multiple_label
        for choice in self.choices:
            label = choice.label
            suffix = label
            stem_len = LABEL_MAX_LEN - (len(suffix) + 3)
            stem = uncut_stem[:stem_len]
            gen_label = ' : '.join((stem, suffix))
            gen_labels.append(stata_string_escape(gen_label))
        return gen_labels

    def render(self) -> str:
        rendered = SPLIT_SELECT_MULTIPLE_UNIT.render(
            orig=self.original_varname,
            padded=self.padded_varname,
            gen_binaries=self.gen_binaries,
            first=self.first,
            last=self.last,
            binary_label=self.binary_label,
        )
        return rendered

    @property
    def first(self) -> str:
        """Get the varname of the first choice option."""
        if not self.gen_binaries:
            return None
        return self.gen_binaries[0].binary_varname

    @property
    def last(self) -> str:
        """Get the varname of the last choice option."""
        if not self.gen_binaries:
            return None
        return self.gen_binaries[-1].binary_varname

    def __repr__(self):
        """Get a representation of this object."""
        return f'<SsmUnit "{self.select_multiple!r}">'


@dataclass(frozen=True)
class VarMethod:
    """A container for a select_multiple and its split method."""
    select_multiple: StataVar
    split_method: str


@dataclass(frozen=True)
class BinaryOptionLabel:
    zero: str
    one: str


@dataclass(frozen=True)
class BinaryDefine:
    label: str
    option_label: BinaryOptionLabel


class SsmDetails:

    def __init__(self, var_methods : List[VarMethod],
                 binary_define: BinaryDefine, split_settings: SplitSettings):
        self.var_methods = var_methods
        self.binary_define = binary_define
        self.split_settings = split_settings

        self.ssm_units = self.get_ssm_units()

    def get_ssm_units(self) -> List[SsmUnit]:
        ssm_units = []
        binary_label = self.binary_define.label
        for var_method in self.var_methods:
            select_multiple = var_method.select_multiple
            split_method = var_method.split_method
            ssm_unit = SsmUnit(select_multiple, split_method, binary_label,
                               self.split_settings)
            ssm_units.append(ssm_unit)
        return ssm_units


DEFAULT_YES_NO = BinaryOptionLabel('No', 'Yes')


VarnameNameLabel = namedtuple('VarnameNameLabel', ['varname', 'name', 'label'])


class SplitSelectMultiple(DoFileSection):

    METHOD_APPEND_NAME = 'append_name'
    METHOD_APPEND_NUMBER = 'append_number'
    METHOD_NAME_ONLY = 'name_only'
    METHOD_NONE = 'none'
    ALLOWED_SPLIT_METHODS = (
        METHOD_APPEND_NAME,
        METHOD_APPEND_NUMBER,
        METHOD_NAME_ONLY,
        METHOD_NONE
    )

    DEFAULT_BINARY_OPTION_LABEL = 'yes_no'
    DEFAULT_BINARY_LABEL = 'o2s_binary_label'

    CHOICE_EXCLUDE_NUMBERS = 'numbers'
    CHOICE_EXCLUDE_NEGATIVE_NUMBERS = 'negative_numbers'
    CHOICE_EXCLUDE_NONE = 'none'
    ALLOWED_CHOICE_EXCLUDES = (
        CHOICE_EXCLUDE_NUMBERS,
        CHOICE_EXCLUDE_NEGATIVE_NUMBERS,
        CHOICE_EXCLUDE_NONE
    )

    DEFAULT_NUMBER_COLUMN = 'o2s_number'

    DEFAULT_SETTINGS = {
        'default_split_method': METHOD_APPEND_NUMBER,
        'binary_option_label': DEFAULT_BINARY_OPTION_LABEL,
        'binary_label': DEFAULT_BINARY_LABEL,
        'choices_to_exclude': CHOICE_EXCLUDE_NEGATIVE_NUMBERS,
        'choice_lists_to_split': [],
        'choice_lists_not_to_split': [],
        'odk_names_to_split': [],
        'odk_names_not_to_split': [],
        'odk_names_to_append_name': [],
        'odk_names_to_append_number': [],
        'odk_names_to_name_only': [],
        'number_column': DEFAULT_NUMBER_COLUMN,
        'strict_numbering': False,
    }

    def __init__(self, dataset: ImportedDataset, settings: dict = None,
                 populate: bool = False):
        self.select_multiples = []
        super().__init__(dataset, settings, populate)

    def populate(self):
        self.select_multiples.clear()
        for var in self.dataset:
            self.analyze_variable(var)

    def analyze_variable(self, var: StataVar):
        if self.should_split(var):
            self.select_multiples.append(var)

    def should_split(self, var: StataVar) -> bool:
        if var.is_dropped():
            return False
        row = var.get_survey_row()
        if row is None:
            return False
        elif not row.is_select_multiple():
            return False
        should_split = True
        if self.default_split_method == self.METHOD_NONE:
            should_split = False
        choice_list = row.choice_list.name
        if choice_list in self.choice_lists_to_split:
            should_split = True
        if choice_list in self.choice_lists_not_to_split:
            should_split = False
        odk_name = var.get_odk_name()
        if odk_name in self.odk_names_to_split:
            should_split = True
        if odk_name in self.odk_names_not_to_split:
            should_split = False
        if odk_name in self.odk_names_to_append_name:
            should_split = True
        if odk_name in self.odk_names_to_append_number:
            should_split = True
        if odk_name in self.odk_names_to_name_only:
            should_split = True
        return should_split

    def do_file_iter(self):
        yield ''

    def get_ssm_details(self) -> SsmDetails:
        var_methods = []
        for select_multiple in self.select_multiples:
            split_method = self.get_split_method(select_multiple)
            if split_method == self.METHOD_NONE:
                continue
            var_method = VarMethod(select_multiple, split_method)
            var_methods.append(var_method)
        binary_define = self.get_binary_define()
        split_settings = SplitSettings(
            self.choices_to_exclude,
            self.which_label,
            self.extra_label,
            self.strict_numbering,
            self.number_column,
        )
        ssm_details = SsmDetails(var_methods, binary_define, split_settings)
        return ssm_details

    def get_split_method(self, var: StataVar) -> str:
        split_method = self.default_split_method
        odk_name = var.get_odk_name()
        if odk_name in self.odk_names_to_append_name:
            split_method = self.METHOD_APPEND_NAME
        if odk_name in self.odk_names_to_append_number:
            split_method = self.METHOD_APPEND_NUMBER
        if odk_name in self.odk_names_to_name_only:
            split_method = self.METHOD_NAME_ONLY
        return split_method

    def get_binary_define(self) -> BinaryDefine:
        binary_label = self.binary_label
        binary_option_label = self.get_binary_option_label()
        binary_define = BinaryDefine(binary_label, binary_option_label)
        return binary_define

    def get_binary_option_label(self) -> BinaryOptionLabel:
        if self.binary_option_label == self.DEFAULT_BINARY_OPTION_LABEL:
            return DEFAULT_YES_NO
        raise ValueError(self.binary_option_label)

    @property
    def default_split_method(self):
        result = self.settings['default_split_method']
        if result not in self.ALLOWED_SPLIT_METHODS:
            methods = (f'"{i}"' for i in self.ALLOWED_SPLIT_METHODS)
            methods = ', '.join(methods)
            msg = ('Under splitting select multiples, "default_split_method" '
                   f'is set to {result}. Please update to be one of {methods}.')
            raise ValueError(msg)
        return result

    @property
    def choices_to_exclude(self):
        result = self.settings['choices_to_exclude']
        if result not in self.ALLOWED_CHOICE_EXCLUDES:
            methods = (f'"{i}"' for i in self.ALLOWED_CHOICE_EXCLUDES)
            methods = ', '.join(methods)
            msg = ('Under splitting select multiples, "choices_to_exclude" '
                   f'is set to {result}. Please update to be one of {methods}.')
            raise ValueError(msg)
        return result

    @property
    def binary_option_label(self):
        result = self.settings['binary_option_label']
        if result != self.DEFAULT_BINARY_OPTION_LABEL:
            msg = ('Under splitting select multiples, "binary_option_label" '
                   'must be '
                   '"yes_no". There are plans to allow other values here.')
            return NotImplementedError(msg)
        return result

    @property
    def binary_label(self):
        result = self.settings['binary_label']
        return result

    @property
    def choice_lists_to_split(self):
        result = self.settings['choice_lists_to_split']
        return result

    @property
    def choice_lists_not_to_split(self):
        result = self.settings['choice_lists_not_to_split']
        return result

    @property
    def odk_names_to_split(self):
        result = self.settings['odk_names_to_split']
        return result

    @property
    def odk_names_not_to_split(self):
        result = self.settings['odk_names_not_to_split']
        return result

    @property
    def odk_names_to_append_name(self):
        result = self.settings['odk_names_to_append_name']
        return result

    @property
    def odk_names_to_append_number(self):
        result = self.settings['odk_names_to_append_number']
        return result

    @property
    def odk_names_to_name_only(self):
        result = self.settings['odk_names_to_name_only']
        return result

    @property
    def number_column(self):
        return self.settings['number_column']

    @property
    def strict_numbering(self):
        """Return true if numbers should strictly come from number_column."""
        return self.settings['strict_numbering']

    def __repr__(self):
        msg = f'<SplitSelectMultiple, size {len(self.select_multiples)}>'
        return msg



"""Module for the command-line interface to odk2stata.

Module functions:
    cli: Run the command-line interface
"""
import argparse
import sys

from .do_file_collection import DoFileCollection


def cli():
    """Run a CLI for this module."""
    if '-V' in sys.argv or '--version' in sys.argv:
        from ..__version__ import __version__
        print(f'odk2stata v{__version__}')
        return

    parser = argparse.ArgumentParser(
        prog='odk2stata',
        description='Generate a configurable do file from an XlsForm.'
    )
    parser.add_argument('xlsform', help='The XlsForm to analyze.')
    parser.add_argument('-s', '--settings',
        help='The settings file. If not supplied then sensible default '
             'settings are used. Access those with '
             '"python3 -m odk2stata.dofile.settings" on the command line.')
    parser.add_argument('-d', '--dataset_source', choices=[
        'briefcase', 'aggregate', 'no_groups'
    ],
        help='Where does the data come from? The option "no_groups" is for '
             'data where the groups have been stripped out. Default is '
             '"briefcase".',
        default='briefcase'
    )
    parser.add_argument('-o', '--outpath', help='Where to save the do file. '
                                                'If not supplied, then the do '
                                                'file is written to STDOUT.')
    parser.add_argument('-V', '--version', action='store_true',
                        help='Print the software version and exit')
    args = parser.parse_args()
    do_file_collection = DoFileCollection.from_file(args.xlsform,
            dataset_source=args.dataset_source, settings_path=args.settings
    )
    if args.outpath:
        do_file_collection.write_out(args.outpath)
        print(f'Saved do file to "{args.outpath}"')
    else:
        print(do_file_collection.render())



"""A module for handling metadata about XLSForm data and Stata."""
import datetime
import getpass
import os.path

from .imported_dataset import ImportedDataset
from ..__version__ import __version__


class Metadata:
    """A class to represent the metadata about data from XLSForms.

    Class attributes:
        DEFAULT_SETTINGS: The default settings for this section
    """

    DEFAULT_SETTINGS = {
        'author': '',
        'timestamp_format': '%Y-%m-%d, %H:%M:%S',
        'case_preserve': False,
        'merge_single_repeat': True,
        'merge_append': True,
        'odk2stata_version': __version__
    }

    def __init__(self, dataset: ImportedDataset, settings: dict = None):
        """Initialize this metadata section.

        Args:
            dataset: The dataset that is produced upon Stata import
            settings: A settings dictionary to update defaults
        """
        self.dataset = dataset
        self.settings = dict(self.DEFAULT_SETTINGS)
        if settings:
            self.settings.update(settings)
        self.odk2stata_version = __version__
        self.odk_source_file = self.dataset.get_odk_source_file()
        self.primary_base = self.get_primary_base()
        self.primary_csv = self.get_primary_csv()
        self.primary_dta = self.get_primary_dta()
        self.secondary_csv = self.get_secondary_csv()
        self.secondary_dta = self.get_secondary_dta()
        self.merge_key = self.get_merge_key()

    def initialize_settings(self, settings: dict):
        if settings:
            self.settings = dict(self.DEFAULT_SETTINGS)
            self.settings.update(settings)

    def update_settings(self, settings: dict):
        if settings:
            self.settings.update(settings)
            new_settings = dict(self.settings)
            self.initialize_settings(new_settings)

    def get_odk_source_file(self):
        full_file = self.dataset.get_odk_source_file()
        _, source_file = os.path.split(full_file)
        return source_file

    def get_primary_base(self):
        filename = self.dataset.primary.dataset_filename
        base, _ = os.path.splitext(filename)
        return base

    def get_primary_csv(self):
        return self.dataset.primary.dataset_filename

    def get_primary_dta(self):
        base = self.get_primary_base()
        filename_dta = f'{base}.dta'
        return filename_dta

    def get_secondary_csv(self):
        if not self.dataset.secondary:
            return None
        return self.dataset.secondary.dataset_filename

    def get_secondary_dta(self):
        if not self.dataset.secondary:
            return None
        csv_filename = self.dataset.secondary.dataset_filename
        path, _ = os.path.splitext(csv_filename)
        filename_dta = f'{path}.dta'
        return filename_dta

    def is_merged_dataset(self):
        return self.dataset.is_merged_dataset()

    def get_merge_key(self):
        if not self.dataset.secondary:
            return None
        merge_column = self.dataset.secondary.begin_repeat
        for var in self.dataset:
            if var.column is merge_column:
                return var.orig_varname

    @property
    def author(self):
        settings_author = self.settings['author']
        author = settings_author if settings_author else getpass.getuser()
        return author

    @property
    def date_created(self):
        now = datetime.datetime.now()
        timestamp = now.strftime(self.settings['timestamp_format'])
        return timestamp

    @property
    def case_preserve(self):
        return self.settings['case_preserve']

    @property
    def merge_single_repeat(self):
        return self.settings['merge_single_repeat']



from typing import Iterator, List, Optional

from .stata_utils import clean_stata_varname, gen_anonymous_varname
from ..dataset.dataset import Dataset
from ..dataset.column import Column
from ..odkform.survey import SurveyRow


class StataVar:

    def __init__(self, column: Column, varname: str):
        self.column = column
        self.orig_varname = varname
        self.varname = varname
        self.dropped = False

    def drop(self) -> None:
        self.dropped = True

    def is_dropped(self) -> bool:
        return self.dropped

    def keep(self) -> None:
        self.dropped = False

    def rename(self, varname: str) -> None:
        self.varname = varname

    def from_single_odk_row(self) -> bool:
        return self.column.from_single_odk_row()

    def based_on_odk_row(self) -> bool:
        return self.column.based_on_odk_row()

    def get_survey_row(self) -> SurveyRow:
        return self.column.get_survey_row()

    def get_odk_type(self) -> str:
        return self.column.get_odk_type()

    def get_odk_name(self) -> str:
        return self.column.get_odk_name()

    def imported_name_is_odk_name(self) -> bool:
        odk_name = self.get_odk_name()
        return odk_name == self.orig_varname

    def set_varname_to_original(self) -> None:
        self.varname = self.orig_varname

    def set_current_varname(self, varname: str) -> None:
        self.varname = varname

    def is_numeric(self) -> bool:
        return self.column.is_numeric()

    def __repr__(self):
        """Get a represenation of this object."""
        return f'StataVar({self.column!r}, {self.orig_varname!r})'


class ImportedDataset:

    def __init__(self, dataset: Dataset, case_preserve: bool,
                 merge_single_repeat: bool, merge_append: bool):
        self.case_preserve = case_preserve
        self.merge_single_repeat = merge_single_repeat
        self.merge_append = merge_append
        self.primary = dataset
        self.secondary = self.get_secondary_dataset()
        self.vars = self.import_vars()

    def get_secondary_dataset(self) -> Optional[Dataset]:
        if not self.merge_single_repeat:
            return None
        to_return = None
        for column in self.primary:
            if column.repeat_dataset:
                if to_return is None:
                    # Keep the first repeat dataset
                    to_return = column.repeat_dataset
                else:
                    # If there are more datasets, return None
                    to_return = None
                    break
        return to_return

    def import_vars(self) -> List[StataVar]:
        primary_vars = self.import_dataset(self.primary, self.case_preserve)
        secondary_vars = self.import_dataset(self.secondary, self.case_preserve)
        combined = []
        if not secondary_vars:
            combined.extend(primary_vars)
        elif not self.merge_append:
            for var in primary_vars:
                combined.append(var)
                if var.column.repeat_dataset:
                    combined.extend(secondary_vars)
        else:
            combined.extend(primary_vars)
            combined.extend(secondary_vars)
        return combined

    @staticmethod
    def import_dataset(dataset, case_preserve) -> List[StataVar]:
        stata_vars = []
        if dataset is None:
            return stata_vars
        column_names = set()
        for i, column in enumerate(dataset, start=1):
            column_name = column.column_name
            stata_cleaned = clean_stata_varname(column_name)
            if not case_preserve:
                stata_cleaned = stata_cleaned.lower()
            stata_varname = stata_cleaned
            if stata_varname in column_names:
                stata_varname = gen_anonymous_varname(i)
            column_names.add(stata_varname)
            this_stata_var = StataVar(column, stata_varname)
            stata_vars.append(this_stata_var)
        return stata_vars

    def is_merged_dataset(self):
        return self.secondary is not None

    def get_odk_source_file(self):
        """Get the ODK source file."""
        return self.primary.odkform.path

    def keep_all(self):
        """Set all variables to be un-dropped."""
        for var in self:
            var.keep()

    def set_all_varnames_to_original(self):
        for var in self:
            var.set_varname_to_original()

    def __iter__(self) -> Iterator[StataVar]:
        """Iterate over the Stata variables in this dataset."""
        return iter(self.vars)

    def __repr__(self):
        """Get a description of this object."""
        return f'<ImportedDataset, len {len(self.vars)}>'



"""Module defining survey building blocks.

Module attributes:
    SurveyRow: Subclassing XlsFormRow, this represents a row in survey
    Survey: The class to define the XlsForm survey
"""
from typing import List, Optional, Tuple

import xlrd.sheet

from .choices import ChoiceList
from .components import XlsFormRow
from .components import Worksheet
from ..error import MismatchedGroupOrRepeatError


class SurveyRow(XlsFormRow):
    """A class to represent a row in the survey tab of an XlsForm.

    This class subclasses XlsFormRow, so some attributes are
    documented there.

    This class has many convenience functions used in other subsystems
    for determining different qualities of the SurveyRow instance.

    Instance attributes:
        ancestors: The list of SurveyRow objects for the groups and
            repeats this row is nested under
        choice_list: The choice_list if this is a choice type question.
            This attribute is set after initialization by OdkForm.
    """

    def __init__(self, rowx: int, row_name: str, row_header: Tuple[str],
                 row_values: list, row_dict: dict,
                 ancestors: list):
        """Initialize a SurveyRow.

        Args:
            rowx: The 0-indexed row in the sheet
            row_name: The value for the name column of this row
            row_header: The headers
            row_values: The values of each cell in this row
            row_dict: An easy lookup to get the value under a header
            ancestors: The list of SurveyRow objects for the groups and
                repeats this row is nested under
        """
        super().__init__(rowx, row_name, row_header, row_values, row_dict)
        self.ancestors = list(ancestors)
        self.choice_list: Optional[ChoiceList] = None

    def get_type(self) -> str:
        """Return the survey row type."""
        return self.row_dict['type']

    def is_select_type(self) -> bool:
        """Return is this survey row is a select type with choices."""
        row_type = self.get_type()
        return row_type.startswith('select')

    def is_select_one(self) -> bool:
        """Return is this a (external) select one type."""
        select_one_starts = (
            'select_one ',
            'select_one_external ',
        )
        row_type = self.get_type()
        return any(row_type.startswith(item) for item in select_one_starts)

    def is_select_multiple(self) -> bool:
        """Return is this a (external) select multiple type."""
        select_multiple_starts = (
            'select_multiple ',
            'select_multiple_external ',
        )
        row_type = self.get_type()
        return any(row_type.startswith(item) for item in select_multiple_starts)

    def is_select_external(self) -> bool:
        """Return is this a select external (single or multiple) type."""
        external_starts = (
            'select_one_external ',
            'select_multiple_external ',
        )
        row_type = self.get_type()
        return any(row_type.startswith(item) for item in external_starts)

    def get_select_list_name(self):
        """Get the list name of the associated choice list."""
        if self.is_select_type():
            row_type = self.get_type()
            _, choice_list = row_type.rsplit(maxsplit=1)
            return choice_list
        return None

    def is_begin_repeat(self):
        """Return is this a begin repeat type."""
        row_type = self.get_type()
        is_begin_repeat = row_type == 'begin repeat'
        return is_begin_repeat

    def is_end_repeat(self):
        """Return is this an end repeat type."""
        row_type = self.get_type()
        is_end_repeat = row_type == 'end repeat'
        return is_end_repeat

    def is_gps(self):
        """Return is this a geopoint type."""
        row_type = self.get_type()
        is_gps = row_type in ('hidden geopoint', 'geopoint')
        return is_gps

    def is_numeric_type(self):
        """Return is this a numeric type."""
        row_type = self.get_type()
        is_numeric = row_type in (
            'hidden decimal',
            'decimal',
            'hidden integer',
            'integer',
            'int',
            'range',
        )
        return is_numeric

    def becomes_column(self):
        """Return if this survey row becomes a column in the dataset."""
        row_type = self.get_type()
        # Note: "begin repeat" does become a column
        non_columns = ('begin group', 'end group', 'end repeat')
        return row_type not in non_columns

    def becomes_single_column(self):
        """Return if this survey row becomes a single column."""
        return self.becomes_column() and not self.is_gps()

    def __hash__(self):
        """Make a hash based on the row number and the row name."""
        return hash((self.rowx, self.row_name))

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<SurveyRow "{self.row_name}" at row {self.rowx}>'
        return msg


class Survey(Worksheet):
    """A class to represent survey in XlsForm.

    Instance attributes:
        header: The header row for the survey tab
        rows: The survey rows in the survey tab
    """

    def __init__(self, sheet: xlrd.sheet.Sheet, datemode: int):
        """Initialize a Survey.

        Args:
            sheet: The xlrd sheet that stores the survey
            datemode: The datemode for the workbook
        """
        self.header: Tuple[str] = self.get_header(sheet, datemode)
        self.rows: List[SurveyRow] = []
        self._parse_survey(sheet, datemode)

    def _parse_survey(self, sheet: xlrd.sheet.Sheet, datemode: int) -> None:
        """Parse the survey tab.

        This method goes through each row in the survey tab. If a row
        has a type and a name, then a SurveyRow is created for it.
        Nesting under groups and repeats is accounted for.

        Args:
            sheet: The xlrd sheet that stores the survey
            datemode: The datemode for the workbook
        """
        ancestors = []
        for i, row in enumerate(sheet.get_rows()):
            if i == 0:
                continue
            row_values = [self.cell_to_value(cell, datemode) for cell in row]
            row_dict = {k: v for k, v in zip(self.header, row_values)}
            row_type = row_dict['type']
            row_name = row_dict['name']
            if row_type in ('end group', 'end repeat'):
                if not ancestors:
                    msg = (f'Found "{row_type}" without matching "begin ..." '
                           f'at row {i+1}')
                    raise MismatchedGroupOrRepeatError(msg)
                ancestors.pop()
            if row_type and row_name:
                survey_row = SurveyRow(i, row_name, self.header, row_values,
                                       row_dict, ancestors)
                self.rows.append(survey_row)
                if row_type in ('begin group', 'begin repeat'):
                    ancestors.append(survey_row)
        if ancestors:
            first_ancestor = ancestors[-1]
            msg = (f'No "end ..." at end of XlsForm to match "begin ..." with '
                   f'name "{first_ancestor}"')
            raise MismatchedGroupOrRepeatError(msg)

    def __getitem__(self, item) -> SurveyRow:
        """Get the survey row specified by the argument."""
        return self.rows[item]

    def __iter__(self):
        """Return an iterator over survey rows."""
        return iter(self.rows)

    def __len__(self):
        """Return the length of survey as the number of survey rows."""
        return len(self.rows)

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<Survey {len(self.rows)} rows, {len(self.header)} columns>'
        return msg



"""A module to describe the dataset based on an ODK file.

This module describes three primary abstractions:

    - DatasetCollection
    - Dataset
    - Column

An ODK file can have repeat groups. When the data are exported to CSV,
then those repeat groups become their own datasets. Therefore, the top
level is the DatasetCollection, which comprises of the primary Dataset
and repeat group Datasets. Each Dataset has Columns.
"""
from .dataset_collection import DatasetCollection





"""Set up the template engine and house template files.

This templates subpackage contains all the template files needed for
creating a do file.
"""
from jinja2 import Environment, PackageLoader


env = Environment(loader=PackageLoader('odk2stata.dofile'))



"""A collection of useful dataset-related functions."""
from enum import Enum
import string
from typing import List


class DatasetSource(Enum):
    """An enumeration of possible dataset sources."""
    BRIEFCASE = 1
    AGGREGATE = 2
    NO_GROUPS = 3

    @classmethod
    def from_string(cls, input: str):
        """Convert a string to enum."""
        if input.lower() == 'briefcase':
            return cls.BRIEFCASE
        elif input.lower() == 'aggregate':
            return cls.AGGREGATE
        elif input.lower() == 'no_groups':
            return cls.NO_GROUPS
        else:
            raise ValueError(input)


def get_column_name(row_name: str, ancestors: List[str],
                    dataset_source: DatasetSource) -> str:
    """Get the column name based on the node hierarchy.

    Args:
        row_name: The name defined in XLSForm.
        ancestors: The names of node ancestors, if any. These are the
            enclosing repeats and begin groups.
        dataset_source: The kind of dataset
    """
    if not isinstance(row_name, str):
        msg = (f'Parameter "row_name" must be of type "str". Got '
               f'{type(row_name)} instead')
        raise TypeError(msg)
    chunks = [*ancestors, row_name]
    if dataset_source == DatasetSource.BRIEFCASE:
        return '-'.join(chunks)
    elif dataset_source == DatasetSource.AGGREGATE:
        return ':'.join(chunks)
    elif dataset_source == DatasetSource.NO_GROUPS:
        return row_name
    else:
        msg = (f'Dataset source "{dataset_source}" should be one of '
               f'{list(DatasetSource)}')
        raise ValueError(msg)


def strip_illegal_chars(text: str) -> str:
    """Remove illegal characters.

    This routine mimics the same behavior in ODK Briefcase when it
    names the resultant dataset file (the .csv output).

    Args:
        text: Input string

    Returns:
        A string with illegal characters removed.
    """
    whitespace_dict = {ord(i): ' ' for i in string.whitespace}
    punctuation_dict = {ord(i): '_' for i in string.punctuation}
    result = text.translate({**whitespace_dict, **punctuation_dict})
    return result



import argparse

from . import SettingsManager


def cli():
    """Run a CLI for this module."""
    prog_desc = 'Generate the default settings ini file.'
    parser = argparse.ArgumentParser(description=prog_desc)
    parser.add_argument('-o', '--outpath', help='Where to save the do file. '
                                                'If not supplied, then the '
                                                'file is written to STDOUT.')
    args = parser.parse_args()
    manager = SettingsManager()

    default_ini = manager.generate_default_ini(args.outpath)
    if default_ini:
        print(default_ini)
    else:
        print(f'Saved do file to "{args.outpath}"')



"""A module for describing datasets and importing them."""
from typing import List

from .column import Column
from .utils import DatasetSource, get_column_name, strip_illegal_chars
from ..odkform import OdkForm
from ..odkform.survey import SurveyRow


class Dataset:
    """A class to describe a Dataset.

    Instance attributes:
        odkform: The OdkForm from whence this dataset comes
        dataset_filename: The filename for this dataset
        dataset_source: The program that created this dataset
        begin_repeat: If this is a repeat group dataset, then this
            attribute is the Column that begins the repeat group
        import_context: An object to keep track of state during import
        columns: A list of Columns contained in this dataset
    """

    def __init__(self, odkform: OdkForm, dataset_source: DatasetSource,
                 begin_repeat: Column = None):
        """Initialize a Dataset.

        Args:
            odkform: The OdkForm from whence this dataset comes
            dataset_filename: The filename for this dataset
            dataset_source: The program that created this dataset
            begin_repeat: If this is a repeat group dataset, then this
                attribute is the Column that begins the repeat group
        """
        self.odkform = odkform
        self.dataset_source = dataset_source
        self.begin_repeat = begin_repeat
        self.dataset_filename = self.get_dataset_filename(
            self.odkform, self.dataset_source, self.begin_repeat
        )
        self.columns: List[Column] = []

        from_briefcase = self.dataset_source == DatasetSource.BRIEFCASE
        if from_briefcase and begin_repeat is None:
            self.start_with_submission_date()

    @staticmethod
    def get_dataset_filename(odkform: OdkForm, dataset_source: DatasetSource,
                             begin_repeat: Column = None) -> str:
        """Get the name of the CSV dataset.

        Args:
            odkform: The source ODK form
            dataset_source: From whence the dataset originates
            begin_repeat: If the dataset is based on a repeat group,
                this is the SurveyRow that begins the repeat group

        Returns:
            The file name, not including any path, of the dataset
        """
        form_title = odkform.settings.form_title
        filename_uncleaned = form_title
        if begin_repeat is not None:
            filename_uncleaned += f'_{begin_repeat.survey_row.row_name}'
        if dataset_source == DatasetSource.AGGREGATE:
            filename_uncleaned += f'_results'
        filename = strip_illegal_chars(filename_uncleaned)
        full_filename = f'{filename}.csv'
        return full_filename

    def start_with_submission_date(self):
        """Start the dataset with a column called SubmissionDate.

        Pre-condition: This should be a dataset with no columns added
        yet.
        """
        self.direct_add('SubmissionDate', None)

    def direct_add(self, column_name: str, survey_row: SurveyRow) -> Column:
        """Add a column to the dataset.

        Args:
            column_name: The name of the column in the dataset
            survey_row: The SurveyRow associated with this column

        Returns:
            The Column object created and added to the datset.
        """
        column = Column(column_name, survey_row)
        self.columns.append(column)
        return column

    def add_next(self, survey_row: SurveyRow) -> List[Column]:
        """Process a SurveyRow and add to this Dataset.

        This method passes the survey row to the ImportContext to
        generate column(s). One survey row almost always corresponds to
        one column. An exception is the geopoint type, which makes four
        columns.

        Args:
            row: The source survey row that becomes a dataset column

        Returns:
            A list of Column objects that have been added to the
            dataset.
        """
        columns = self.get_next(survey_row)
        self.columns.extend(columns)
        return columns

    def get_next(self, survey_row: SurveyRow) -> List[Column]:
        """Get the next columns based on the input SurveyRow.

        This method contains the business logic of processing the next
        survey row. It keeps track of Stata varnames, column number,
        and repeat group datasets.

        Args:
            survey_row: The next survey row to process

        Returns:
            A list of Columns associated with this survey row
        """
        columns = []
        row_name = survey_row.row_name
        ancestors = self.get_ancestors(survey_row)
        column_name = get_column_name(row_name, ancestors, self.dataset_source)
        if survey_row.is_gps():
            for suffix in 'Latitude', 'Longitude', 'Altitude', 'Accuracy':
                next_column_name = f'{column_name}-{suffix}'
                next_column = Column(next_column_name, survey_row)
                columns.append(next_column)
        elif survey_row.is_begin_repeat():
            # TODO: write a test if the repeat is inside a group.
            # TODO: is it the fully-qualified name? or just the row name?
            next_column_name = f'SET-OF-{column_name}'
            next_column = Column(next_column_name, survey_row)
            columns.append(next_column)
        else:
            next_column = Column(column_name, survey_row)
            columns.append(next_column)
        return columns

    def get_ancestors(self, survey_row: SurveyRow) -> List[str]:
        """Get ancestors for the SurveyRow.

        For naming purposes, we need to know the ancestors of rows
        within repeats or groups. If a row is inside a begin repeat,
        then we only take ancestors after that begin repeat.

        Args:
            survey_row: The row for which to get ancestors

        Returns:
            A list of the names of the ancestor nodes.
        """
        ancestors = list(survey_row.ancestors)
        if self.begin_repeat is not None:
            while ancestors:
                if self.begin_repeat.survey_row is ancestors.pop(0):
                    break
        return [i.row_name for i in ancestors]

    def get_datasets(self) -> list:
        """Return this dataset and repeat datasets under it.

        This is a depth-first search.

        Returns:
            A list of Datasets. This dataset is last in the list.
        """
        datasets = []
        for column in self:
            if column.repeat_dataset is not None:
                repeats = column.repeat_dataset.get_datasets()
                datasets.extend(repeats)
        datasets.append(self)
        return datasets

    def appended_iter(self):
        """Iterate over the columns in appended dataset order.

        This a recursive method. It descends into associated repeat
        groups as it goes along.

        1. Iterate through all columns of this dataset
        2. Along the way, keep track of the associated repeat datasets
        3. Descend and recurse in those discovered datasets

        Yields:
            The next Column
        """
        repeat_datasets = []
        for column in self:
            yield column
            if column.repeat_dataset is not None:
                repeat_datasets.append(column.repeat_dataset)
        for repeat_dataset in repeat_datasets:
            yield from repeat_dataset.appended_iter()

    def inserted_iter(self):
        """Iterate over the columns as in the original ODK ordering.

        This a recursive method. It descends into associated repeat
        groups as it goes along.

        Yields:
            The next Column.
        """
        for column in self:
            yield column
            if column.repeat_dataset is not None:
                yield from column.repeat_dataset.inserted_iter()

    def is_repeat_dataset(self) -> bool:
        """Return if this dataset is a repeat dataset."""
        return self.begin_repeat is not None

    def __getitem__(self, key):
        """Return column specified by key."""
        return self.columns[key]

    def __len__(self):
        """Return the number of columns."""
        return len(self.columns)

    def __iter__(self):
        """Return an iterator over the columns."""
        return iter(self.columns)

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<Dataset with {len(self.columns)} columns>'
        return msg



"""A module to handle everything related to ODK choices.

Module attributes:
    NumberNameChoice: A namedtuple to encapsulate the returned data for
        choice numberings.
    Choices: A class to handle all choices together. Choices can come
        from disparate source data
    ChoiceListTab: A class to represent the choices sheet, be it
        "choices" or "external_choices"
    ChoiceList: A single choice list found at a choices source data
"""
from collections import defaultdict, namedtuple
from typing import List, Dict, Tuple

import xlrd

from .components import XlsFormRow
from .components import Worksheet


NumberNameChoice = namedtuple('NumberNameChoice', ('number', 'name', 'choice'))


class ChoiceList:
    """A class to represent a single choice list.

    This class is meant to be read-only after initialization.

    Instance attributes:
        name: The name of the choice list
        choices: The list of choices
        sheet_name: The sheet name where these choices came from
        row_header: Calculated from the first row of the choices.
    """

    def __init__(self, name: str, choices: List[XlsFormRow], sheet_name: str):
        """Initialize a ChoiceList.

        Also stores the row_header for the choices as determined by
        the first choice row.

        Args:
            name: The name of the choice list
            choices: The list of choices
            sheet_name: The sheet name where these choices came from
        """
        self.name = name
        self.choices = choices
        self.sheet_name = sheet_name
        self.row_header = self.choices[0].row_header

    def are_choice_names_all_integer(self) -> bool:
        """Determine if all choice options have an integer ODK name."""
        return all(isinstance(i.row_name, int) for i in self)

    def get_choices_flexibly_numbered(self, extra_number: str = None) \
            -> List[NumberNameChoice]:
        """Get the choice list with flexible option numbers.

        Args:
            extra_number: A column where to look for a number. A number
                found here takes precedence.

        Returns:
            A list of tuples. Each tuple corresponds to a choice option
            and has the assigned choice option number, the choice name,
            and the XlsFormRow representing the choice row.
        """
        result = []
        next_number = 1
        for choice in self.choices:
            name = choice.row_name
            extra_found = choice.row_dict.get(extra_number)
            if isinstance(extra_found, int):
                value = extra_found
            elif isinstance(name, int):
                value = name
            else:
                value = next_number
            next_number = max(next_number, value) + 1
            result.append(NumberNameChoice(value, str(name), choice))
        return result

    def get_choices_strictly_numbered(self, number_column: str ='name') \
            -> List[NumberNameChoice]:
        """Get the choice list with strict option numbers.

        Args:
            number_column: A column where to look for a number.

        Returns:
            A list of tuples. Each tuple corresponds to a choice option
            and has the assigned choice option number, the choice name,
            and the XlsFormRow representing the choice row.

        Raises:
            KeyError if the supplied number_column is not in the
            header, or ValueError if not all entries are integer.
        """
        if number_column not in self.row_header:
            msg = (f'Unable to find "{number_column}" in column headers '
                   f'for choice list "{self.name}"')
            raise KeyError(msg)
        numbers = [row.row_dict[number_column] for row in self.choices]
        if any(not isinstance(i, int) for i in numbers):
            msg = (f'Choice list "{self.name}" does not define all '
                   f'options to have a number in the "{number_column}" '
                   f'column')
            raise ValueError(msg)
        result = []
        for number, choice in zip(numbers, self.choices):
            result.append(
                NumberNameChoice(number, str(choice.row_name), choice)
            )
        return result

    def __hash__(self):
        """Make a hash based on the sheet name and the list name."""
        return hash((self.sheet_name, self.name))

    def __iter__(self):
        """Return an iterator over individual choices in this list."""
        return iter(self.choices)

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<ChoiceList "{self.name}">'
        return msg


class ChoiceListTab(Worksheet):
    """A class to represent a sheet of choices or external choices.

    Instance attributes:
        header: The header for the tab
        choices: A dictionary of choice names and ChoiceLists
    """

    def __init__(self, sheet: xlrd.sheet.Sheet, datemode: int):
        """Initialize a ChoiceListTab object.

        Args:
            sheet: The xlrd sheet object for this sheet
            datemode: The xlrd datemode for the workbook
        """
        self.header: Tuple[str] = None
        self.choices: Dict[str, ChoiceList] = {}
        self.build_choices(sheet, datemode)

    def build_choices(self, sheet: xlrd.sheet.Sheet, datemode: int) -> None:
        """Parse the tab of ODK choices.

        This function modifies the instance attributes `header` and
        `choices` if there is something stored in the tab of ODK
        chocies.

        Args:
            sheet: The xlrd sheet object for this sheet
            datemode: The xlrd datemode for the workbook
        """
        _choices_dict = defaultdict(list)
        if sheet is not None:
            try:
                self.header = self.get_header(sheet, datemode)
                for i, row in enumerate(sheet.get_rows()):
                    if i == 0:
                        continue
                    row_values = [self.cell_to_value(cell, datemode) for cell in
                                  row]
                    row_dict = {k: v for k, v in zip(self.header, row_values)}
                    row_list_name = row_dict['list_name']
                    row_name = row_dict['name']
                    if row_list_name and row_name:
                        choice_row = XlsFormRow(i, row_name, self.header,
                                                row_values, row_dict)
                        _choices_dict[row_list_name].append(choice_row)
            except IndexError:
                # No header row found. Then no choices.
                pass
        for name, choices in _choices_dict.items():
            choice_list = ChoiceList(name, choices, sheet.name)
            self.choices[name] = choice_list

    def __len__(self):
        """Return the number of choice list names in this object."""
        return len(self.choices)

    def __iter__(self):
        """Return an iterator over the choice list names."""
        return iter(self.choices)

    def __getitem__(self, item):
        """Get the ChoiceList identified by the argument."""
        choice_list = self.choices.get(item)
        if choice_list is None:
            raise KeyError(item)
        return choice_list

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<ChoiceListTab with {len(self)} choice lists>'
        return msg


class Choices(Worksheet):
    """A class to handle all choices together for the OdkForm.

    This class is the container for all choices used by the OdkForm.

    Instance attributes:
        choices: A tab with choice lists
        external_choices: A tab with external choice lists
    """

    def __init__(self, workbook: xlrd.Book):
        """Initialize the choices object.

        Args:
            workbook: The xlrd book object
        """
        self.choices = self.parse_choices_from_sheet(workbook, 'choices')
        self.external_choices = \
            self.parse_choices_from_sheet(workbook, 'external_choices')

    @staticmethod
    def parse_choices_from_sheet(workbook: xlrd.Book, sheet_name: str) -> ChoiceListTab:
        """Parse choices from a choice sheet.

        Args:
            workbook: The xlrd book object
            sheet_name: The sheet name to parse as a choice sheet

        Returns:
            A ChoiceListTab representing the specified sheet
        """
        sheet = None
        try:
            sheet = workbook.sheet_by_name(sheet_name)
        except xlrd.biffh.XLRDError:
            pass
        return ChoiceListTab(sheet, workbook.datemode)

    def __repr__(self):
        """Get a representation of this object."""
        msg = (f'<Choices with {len(self.choices)} choice lists, '
               f'{len(self.external_choices)} external choice lists>')
        return msg



"""A module for handling datasets derived from ODK files."""
from typing import List

from .dataset import Dataset
from .utils import DatasetSource
from ..odkform import OdkForm


class DatasetCollection:
    """A class to represent all data associated with an ODK form.

    Instance attributes:
        odkform: The source ODK form
        dataset_source: From whence the dataset originates
        primary: The primary dataset. This always exists
    """

    def __init__(self, odkform: OdkForm, dataset_source: DatasetSource):
        """Initialize a DatasetCollection.

        Right now, support multiple, but not nested repeats. Therefore,
        repeats are stored as a list.

        Args:
            odkform: The source ODK form
            dataset_source: From whence the dataset originates
        """
        self.odkform = odkform
        self.dataset_source = dataset_source
        self.primary = self.odkform_to_dataset(odkform, dataset_source)

    @staticmethod
    def odkform_to_dataset(odkform: OdkForm, dataset_source: DatasetSource) \
            -> Dataset:
        """Convert an OdkForm into a Dataset.

        Args:
            odkform: The source ODK form
            dataset_source: From whence the dataset originates

        Returns:
            A Dataset representing the primary dataset
        """
        dataset_stack = []
        primary_dataset = Dataset(odkform, dataset_source)
        current_dataset = primary_dataset
        for row in odkform.survey:
            if row.becomes_column():
                new_columns = current_dataset.add_next(row)
                if row.is_begin_repeat():
                    dataset_stack.append(current_dataset)
                    new_column = new_columns[0]
                    new_column.repeat_dataset = Dataset(odkform, dataset_source,
                                                        begin_repeat=new_column)
                    current_dataset = new_column.repeat_dataset
            elif row.is_end_repeat():
                current_dataset = dataset_stack.pop()
        return primary_dataset

    def get_repeat_datasets(self) -> List[Dataset]:
        """Return the repeat Datasets in this collection."""
        all_datasets = self.get_datasets()
        return all_datasets[:-1]

    def get_datasets(self) -> List[Dataset]:
        """Return all Datasets in this collection.

        This includes the primary dataset as the last element.

        Returns:
            A list of datasets in this collection.
        """
        return self.primary.get_datasets()

    def can_merge_single_repeat(self) -> bool:
        """Return if there a single repeat.

        We can merge a single repeat if there are exactly two datasets.
        """
        return len(self.get_datasets()) == 2

    def merged_iter(self):
        """Iterate over the columns in merged dataset order.

        Yields:
            The next Column
        """
        yield from self.primary.merged_iter()

    def ordered_iter(self):
        """Iterate over the columns as in the original ODK ordering.

        Yields:
            The next Column
        """
        yield from self.primary.ordered_iter()

    @classmethod
    def from_file(cls, path: str, dataset_source: str):
        """Initialize a DatasetCollection with a filename.

        This method is provided to initialize a DatasetCollection
        easily, possibly from the command-line or REPL.

        Args:
            path: The path where to find the ODK file
            dataset_source: From whence the dataset originates. This
                must be a string that DatasetSource understands.

        Returns:
            An initialized DatasetCollection
        """
        source = DatasetSource.from_string(dataset_source)
        odkform = OdkForm(path)
        return cls(odkform, source)

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'DatasetCollection({self.odkform!r}, "{self.dataset_source}")'
        return msg



"""Generate do files for ODK using Python.

There are three primary abstractions in odk2stata. They are

    - odkform
    - dataset
    - dofile

The order is important because an XlsForm is created first as an ODK
form. That ODK form determines what the dataset looks like. The dataset
is loaded into Stata and a do file analyzes the dataset.
"""



from collections import defaultdict, namedtuple
from typing import List
import textwrap

from .do_file_section import DoFileSection
from .imported_dataset import ImportedDataset, StataVar
from .stata_utils import (get_varname_comments, is_valid_stata_varname,
                          make_invalid_varname_comment, safe_stata_string_quote,
                          stata_string_escape)
from .templates import env
from ..odkform.choices import ChoiceList


class EncodeSelectOne(DoFileSection):

    DEFAULT_SETTINGS = {
        'encode_select_ones': True,
        'encode_external_select_ones': False,
        'odk_names_to_encode': [],
        'odk_names_not_to_encode': [],
        'choice_lists_not_to_encode': [],
        'number_column': 'o2s_number',
        'strict_numbering': False,
        'label_replace_column': 'first_label',
    }

    def __init__(self, dataset: ImportedDataset, settings: dict,
                 populate: bool = False):
        self.select_ones: List[StataVar] = []
        super().__init__(dataset, settings, populate)

    def populate(self):
        self.select_ones.clear()
        for var in self.dataset:
            self.analyze_variable(var)

    def analyze_variable(self, var: StataVar):
        if self.should_encode(var):
            self.select_ones.append(var)

    def should_encode(self, var: StataVar):
        survey_row = var.column.survey_row
        if survey_row is None:
            return False
        if not survey_row.is_select_one():
            return False
        if var.is_dropped():
            return False
        should_encode = self.encode_select_ones
        if survey_row.is_select_external():
            should_encode = self.encode_external_select_ones
        if survey_row.choice_list.name in self.choice_lists_not_to_encode:
            should_encode = False
        if var.get_odk_name() in self.odk_names_to_encode:
            should_encode = True
        if var.get_odk_name() in self.odk_names_not_to_encode:
            should_encode = False
        return should_encode

    def get_encode_details(self):
        details = EncodeDetails(
            self.select_ones,
            self.number_column,
            self.strict_numbering,
            self.which_label,
            self.extra_label,
        )
        return details

    def do_file_iter(self):
        # TODO: I don't like this method of getting the do-code
        details = self.get_encode_details()
        yield details.get_encode_select_one_do()

    @property
    def encode_select_ones(self):
        return self.settings['encode_select_ones']

    @property
    def encode_external_select_ones(self):
        return self.settings['encode_external_select_ones']

    @property
    def odk_names_to_encode(self):
        return self.settings['odk_names_to_encode']

    @property
    def odk_names_not_to_encode(self):
        return self.settings['odk_names_not_to_encode']

    @property
    def choice_lists_not_to_encode(self):
        return self.settings['choice_lists_not_to_encode']

    @property
    def number_column(self):
        return self.settings['number_column']

    @property
    def strict_numbering(self):
        """Return true if numbers should strictly come from number_column."""
        return self.settings['strict_numbering']

    def __repr__(self):
        """Get a representation of this object."""
        msg = f'<EncodeSelectOne, size {len(self.select_ones)}>'
        return msg


class EncodeDetails:

    def __init__(self, select_ones: List[StataVar], number_column: str,
                 strict_numbering: bool, which_label: str, extra_label: str):
        self.select_ones = select_ones
        choice_lists = self.get_unique_choice_lists()
        sorted_choice_lists = sorted(list(choice_lists), key=lambda x: x.name)
        encode_choice_list_lookup = {}
        self.encode_choice_lists = []
        for choice_list in sorted_choice_lists:
            encode_choice_list = EncodeChoiceList(
                choice_list,
                number_column,
                strict_numbering,
                which_label,
                extra_label
            )
            self.encode_choice_lists.append(encode_choice_list)
            encode_choice_list_lookup[choice_list] = encode_choice_list
        self.encode_singleton = []
        sorted_select_ones = defaultdict(list)
        for select_one in self.select_ones:
            choice_list = select_one.column.survey_row.choice_list
            if len(select_one.varname) > 30:
                encode_choice_list = encode_choice_list_lookup[choice_list]
                encode_singleton = EncodeSingleton(select_one,
                                                   encode_choice_list)
                self.encode_singleton.append(encode_singleton)
            else:
                sorted_select_ones[choice_list].append(select_one)
        self.encode_for = []
        for key, value in sorted_select_ones.items():
            encode_choice_list = encode_choice_list_lookup[key]
            if len(value) > 1:
                encode_for = EncodeFor(value, encode_choice_list)
                self.encode_for.append(encode_for)
            else:
                encode_singleton = EncodeSingleton(value[0], encode_choice_list)
                self.encode_singleton.append(encode_singleton)

    def get_unique_choice_lists(self):
        unique = set()
        for var in self.select_ones:
            choice_list = var.column.survey_row.choice_list
            unique.add(choice_list)
        return unique

    def get_encode_select_one_do(self):
        label_define = '\n'.join((
            item.get_label_define_do() for item in self.encode_choice_lists
        ))
        singletons = '\n\n'.join((
            item.get_singleton_do() for item in self.encode_singleton
        ))
        foreach = '\n\n'.join((
            item.get_for_do() for item in self.encode_for
        ))
        label_replace = '\n'.join((
            item.get_label_replace_do() for item in self.encode_choice_lists
        ))
        return '\n\n'.join([label_define, singletons, foreach, label_replace])


LabelDefineOption = namedtuple('LabelDefineOption', ['number', 'label'])
LABEL_DEFINE_UNIT = env.get_template('label_define_unit.do')


class EncodeChoiceList:

    def __init__(self, choice_list: ChoiceList, number_column: str,
                 strict_numbering: bool, which_label: str, extra_label: str):
        self.choice_list = choice_list
        self.list_name = self.get_list_name()
        self.number_column = number_column
        self.strict_numbering = strict_numbering
        self.which_label = which_label
        self.extra_label = extra_label

    def get_list_name(self):
        # TODO: self.choice_list.name may not be valid
        return self.choice_list.name

    def get_label_define_do(self) -> str:
        list_varname = self.list_name
        label_defines = self.get_label_define_options()
        full_label_define = LABEL_DEFINE_UNIT.render(
            varname=list_varname,
            options_list=label_defines,
            replace=False
        )
        return full_label_define

    def get_label_replace_do(self) -> str:
        list_varname = self.list_name
        label_defines = self.get_label_replace_options()
        full_label_define = LABEL_DEFINE_UNIT.render(
            varname=list_varname,
            options_list=label_defines,
            replace=True
        )
        return full_label_define

    def get_label_define_options(self):
        result = []
        numbers = self.choice_list_numbers()
        labels = self.choice_list_names()
        for number, label in zip(numbers, labels):
            label_define = LabelDefineOption(number, label)
            result.append(label_define)
        return result

    def get_label_replace_options(self):
        result = []
        numbers = self.choice_list_numbers()
        labels = self.choice_list_labels()
        for number, label in zip(numbers, labels):
            label_define = LabelDefineOption(number, label)
            result.append(label_define)
        return result

    def choice_list_numbers(self):
        if self.strict_numbering:
            return self.choice_list_numbers_strict()
        return self.choice_list_numbers_flexible()

    def choice_list_numbers_strict(self):
        if not self.number_column:
            msg = ('With strict choice numbering, a "number_column" must be '
                   'specified in settings')
            raise ValueError(msg)
        if self.number_column not in self.choice_list.row_header:
            msg = (f'Unable to find "{self.number_column}" in column headers '
                   f'for choice list "{choice_list.name}"')
            raise ValueError(msg)
        numbers = [row.row_dict[self.number_column] for row in self.choice_list]
        if any(not isinstance(i, int) for i in numbers):
            msg = (f'Choice list "{self.choice_list.name}" does not define all '
                   f'options to have a number in the "{self.number_column}" '
                   f'column')
            raise ValueError(msg)
        result = [str(i) for i in numbers]
        return result

    def choice_list_numbers_flexible(self):
        result = []
        next_number = 1
        for choice in self.choice_list:
            name = choice.row_dict['name']
            number_column = choice.row_dict.get(self.number_column)
            if isinstance(number_column, int):
                value = number_column
            elif isinstance(name, int):
                value = name
            else:
                value = next_number
            result.append(str(value))
            next_number = max(next_number, value) + 1
        return result

    def choice_list_names(self):
        names = (str(choice.row_name) for choice in self.choice_list)
        result = [safe_stata_string_quote(name) for name in names]
        return result

    def choice_list_labels(self):
        result = []
        for choice in self.choice_list:
            which_label = self.which_label
            if which_label == 'first_label':
                header = choice.row_header
                first_label = next(i for i in header if i.startswith('label'))
                label = choice.row_dict[first_label]
            elif which_label in choice.row_dict:
                label = choice.row_dict[which_label]
            else:
                msg = ('Encode select one should specify a value for '
                       '"label_replace_column" that is either "first_label", '
                       'or a column header in the choices/external_choices '
                       'tab. Instead, it is set to '
                       f'"{label_replace_column}".')
                raise ValueError(msg)
            final_label = str(label)
            extra_label = choice.row_dict.get(self.extra_label, '')
            if extra_label != '':
                final_label = str(extra_label)
            escaped = stata_string_escape(final_label)
            result.append(escaped)
        return result


ENCODE_SELECT_ONE_UNIT = env.get_template('encode_select_one_unit.do')


class EncodeSingleton:

    def __init__(self, select_one: StataVar,
                 encode_choice_list: EncodeChoiceList):
        self.select_one = select_one
        self.encode_choice_list = encode_choice_list

    def get_singleton_do(self) -> str:
        orig = self.select_one.varname
        gen = f'{orig[:30]}V2'
        lab = self.encode_choice_list.list_name
        singleton_do = ENCODE_SELECT_ONE_UNIT.render(
            orig=orig,
            gen=gen,
            lab=lab,
        )
        return singleton_do


ENCODE_SELECT_ONE_FOR_UNIT = env.get_template('encode_select_one_for_unit.do')


class EncodeFor:

    STATA_FOR_VAR = 'var'
    SUFFIX = 'V2'
    WIDTH = 120

    def __init__(self, select_ones: List[StataVar],
                 encode_choice_list: EncodeChoiceList):
        self.select_ones = select_ones
        self.encode_choice_list = encode_choice_list

    def get_foreach_line(self, stata_var, width=WIDTH):
        intro = f'foreach {stata_var} in '
        intro_width = len(intro)
        wrap_width = width - intro_width
        joined_varnames = ' '.join((var.varname for var in self.select_ones))
        subsequent = ' ' * intro_width
        wrapped = textwrap.wrap(joined_varnames, width=wrap_width,
                                initial_indent=intro,
                                subsequent_indent=subsequent)
        joined_wrapped = ' ///\n'.join(wrapped)
        complete_text = f'{joined_wrapped} '
        return complete_text

    def get_for_do(self, stata_var=STATA_FOR_VAR, suffix=SUFFIX) -> str:
        foreach = self.get_foreach_line(stata_var)
        lab = self.encode_choice_list.list_name
        for_do = ENCODE_SELECT_ONE_FOR_UNIT.render(
            foreach=foreach,
            lab=lab,
            var=stata_var,
            suffix=suffix,
        )
        return for_do



from typing import List

from .do_file_section import DoFileSection
from .imported_dataset import ImportedDataset, StataVar


class Destring(DoFileSection):

    DEFAULT_SETTINGS = {
        'odk_names_to_destring': [],
    }

    def __init__(self, dataset: ImportedDataset, settings: dict = None,
                 populate: bool = False):
        self.destring: List[StataVar] = []
        super().__init__(dataset, settings, populate)

    def populate(self):
        self.destring.clear()
        for var in self.dataset:
            self.analyze_variable(var)

    def analyze_variable(self, var: StataVar):
        if self.should_destring(var):
            self.destring.append(var)

    def should_destring(self, var: StataVar):
        if var.is_dropped():
            return False
        should_destring = False
        if var.is_numeric():
            should_destring = True
        elif var.get_odk_name() in self.odk_names_to_destring:
            should_destring = True
        return should_destring

    def destring_vars_iter(self):
        return (var.varname for var in self.destring)

    def do_file_iter(self):
        for column in self.destring:
            yield self.destring_do(column)

    def destring_do(self, var: StataVar) -> str:
        varname = var.varname
        destring = f'destring {varname}, replace'
        return destring

    @property
    def odk_names_to_destring(self):
        result = self.settings['odk_names_to_destring']
        return result





import os.path
import threading

import wx

from ..dofile.do_file_collection import DoFileCollection


EVT_COMPLETE_ID = wx.NewId()


def evt_complete(win, func):
    win.Connect(-1, -1, EVT_COMPLETE_ID, func)


class CompleteEvent(wx.PyEvent):

    def __init__(self, message, success):

        super().__init__()
        self.SetEventType(EVT_COMPLETE_ID)
        self.message = message
        self.success = success


class Worker(threading.Thread):

    def __init__(self, panel, xlsform_path, settings_path, output_dir):
        super().__init__()

        self.panel = panel
        self.xlsform_path = xlsform_path
        self.settings_path = settings_path
        self.filename = os.path.splitext(os.path.basename(xlsform_path))[0] + '.do'
        self.output_path = os.path.join(output_dir, self.filename)

    def run(self):
        do_files = DoFileCollection.from_file(self.xlsform_path,
                                              settings_path=self.settings_path)
        do_files.write_out(self.output_path)
        message = f'Do file saved to "{self.output_path}"\n'
        wx.PostEvent(self.panel, CompleteEvent(message, True))



#!/usr/bin/env python3
import datetime
import os.path
import webbrowser

import wx

from .worker import Worker, evt_complete


TITLE = 'odk2stata'
APP_QUIT = 1
APP_ABOUT = 2
MAIN_WINDOW_WIDTH = 550
MAIN_WINDOW_HEIGHT = 525
MAX_PATH_LENGTH = 45


class MainWindow(wx.Frame):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.xlsform_path = ''
        self.settings_path = ''
        self.output_path = ''
        self.messages = ''

        self.init_ui()
        self.init_menu()
        self.Center()

        self.Bind(wx.EVT_CLOSE, self.on_quit)

    def init_ui(self):

        panel = wx.Panel(self)

        grid = wx.GridBagSizer(5, 5)

        grid.Add((-1,1), pos=(0, 0), flag=wx.TOP | wx.LEFT | wx.BOTTOM,
                  border=15)

        xlsform_st = wx.StaticText(panel, label="XLS Form")
        grid.Add(xlsform_st, pos=(1, 0), flag=wx.LEFT, border=10)

        self.xlsform_file_tc = wx.TextCtrl(panel)
        self.xlsform_file_tc.SetEditable(False)
        grid.Add(self.xlsform_file_tc, pos=(1, 1), span=(1, 3),
                  flag=wx.EXPAND)

        self.xlsform_btn = wx.Button(panel, label="Browse...")
        grid.Add(self.xlsform_btn, pos=(1, 4), flag=wx.RIGHT, border=10)
        self.xlsform_btn.Bind(wx.EVT_BUTTON, self.on_xlsform_btn)

        settings_st = wx.StaticText(panel, label="Settings")
        grid.Add(settings_st, pos=(2, 0), flag=wx.LEFT, border=10)

        self.settings_file_tc = wx.TextCtrl(panel)
        self.settings_file_tc.SetEditable(False)
        grid.Add(self.settings_file_tc, pos=(2, 1), span=(1, 3), flag=wx.EXPAND)
        self.settings_file_tc.SetLabel('(use program defaults)')

        self.settings_btn = wx.Button(panel, label="Browse...")
        grid.Add(self.settings_btn, pos=(2, 4), flag=wx.RIGHT, border=10)
        self.settings_btn.Bind(wx.EVT_BUTTON, self.on_settings_btn)

        line = wx.StaticLine(panel)
        grid.Add(line, pos=(3, 0), span=(1, 5),
                 flag=wx.EXPAND | wx.BOTTOM, border=10)

        output_st = wx.StaticText(panel, label="Output")
        grid.Add(output_st, pos=(4, 0), flag=wx.LEFT, border=10)

        self.output_file_tc = wx.TextCtrl(panel)
        self.output_file_tc.SetEditable(False)
        grid.Add(self.output_file_tc, pos=(4, 1), span=(1, 3), flag=wx.EXPAND)

        self.output_btn = wx.Button(panel, label="Browse...")
        grid.Add(self.output_btn, pos=(4, 4), flag=wx.RIGHT, border=10)
        self.output_btn.Bind(wx.EVT_BUTTON, self.on_output_btn)

        messages_sb = wx.StaticBox(panel, label="Messages")

        boxsizer = wx.StaticBoxSizer(messages_sb, wx.VERTICAL)
        self.messages_tc = wx.TextCtrl(panel, style=wx.TE_MULTILINE, size=(-1, 250))
        self.messages_tc.SetEditable(False)
        self.messages_tc.SetValue(self.messages)
        boxsizer.Add(self.messages_tc, proportion=1, flag=wx.EXPAND)

        grid.Add(boxsizer, pos=(5, 0), span=(2, 5),
                  flag=wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, border=10)

        self.help_btn = wx.Button(panel, label='Help')
        grid.Add(self.help_btn, pos=(7, 0), flag=wx.LEFT, border=10)
        self.help_btn.Bind(wx.EVT_BUTTON, self.on_help_btn)

        self.run_btn = wx.Button(panel, label="Run")
        grid.Add(self.run_btn, pos=(7, 4), flag=wx.BOTTOM | wx.RIGHT, border=10)
        self.run_btn.Bind(wx.EVT_BUTTON, self.on_run_btn)
        self.run_btn.Disable()

        grid.AddGrowableCol(2)

        panel.SetSizer(grid)
        # grid.Fit(self)

        evt_complete(self, self.on_worker_complete)

    def on_xlsform_btn(self, e):
        dlg = wx.FileDialog(
            self, message='Choose a file',
            defaultFile='',
            wildcard='XLSForm files (*.xls,*.xlsx)|*.xls;*.xlsx',
            style=wx.FD_OPEN | wx.FD_CHANGE_DIR
        )
        dlg.CentreOnParent()
        if dlg.ShowModal() == wx.ID_OK:
            self.xlsform_path = dlg.GetPath()
            self.xlsform_file_tc.SetLabel(
                self.shorten_string(self.xlsform_path, MAX_PATH_LENGTH))
            self.output_path = os.path.dirname(self.xlsform_path)
            self.output_file_tc.SetLabel(
                self.shorten_string(self.output_path, MAX_PATH_LENGTH))
            if self.xlsform_path and self.output_path:
                self.run_btn.Enable()
            else:
                self.run_btn.Disable()
        dlg.Destroy()

    def on_settings_btn(self, e):
        dlg = wx.FileDialog(
            self, message='Choose a file',
            defaultFile='',
            wildcard='Settings file (*.ini,*.cfg)|*.ini;*.cfg',
            style=wx.FD_OPEN | wx.FD_CHANGE_DIR
        )
        dlg.CentreOnParent()
        if dlg.ShowModal() == wx.ID_OK:
            self.settings_path = dlg.GetPath()
            self.settings_file_tc.SetLabel(
                self.shorten_string(self.settings_path, MAX_PATH_LENGTH))
        dlg.Destroy()

    def on_output_btn(self, e):
        dlg = wx.DirDialog(
            self, message='Choose a location',
            defaultPath=self.output_path,
            style=wx.DD_DEFAULT_STYLE | wx.DD_CHANGE_DIR)
        dlg.CentreOnParent()
        if dlg.ShowModal() == wx.ID_OK:
            self.output_path = dlg.GetPath()
            self.output_file_tc.SetLabel(
                self.shorten_string(self.output_path, MAX_PATH_LENGTH))
            if self.xlsform_path and self.output_path:
                self.run_btn.Enable()
            else:
                self.run_btn.Disable()
        dlg.Destroy()

    def on_run_btn(self, e):
        self.enable_ui(False)
        dt_now = datetime.datetime.now()
        now = dt_now.strftime("%H:%M:%S")
        start_msg = f'[Begin conversion at {now}]\n'
        xlsform_msg = f'XLS Form: {os.path.basename(self.xlsform_path)}\n'
        settings_basename = os.path.basename(
            self.settings_path) if self.settings_path else 'Using default settings'
        settings_msg = 'Settings: {}\n'.format(settings_basename)

        self.messages_tc.AppendText(start_msg)
        self.messages_tc.AppendText(xlsform_msg)
        self.messages_tc.AppendText(settings_msg)

        result_thread = Worker(self, self.xlsform_path, self.settings_path, self.output_path)
        result_thread.start()

    def on_help_btn(self, e):
        webbrowser.open("https://www.github.com/jkpr/odk2stata")

    def on_worker_complete(self, event):
        if event.success:
            self.messages_tc.AppendText('\nSUCCESS!\n')
        else:
            self.messages_tc.AppendText('\nSomething went wrong\n')
        self.messages_tc.AppendText(event.message)
        self.messages_tc.AppendText('- - - - - - - - - - - - - - - - - - -\n')
        self.enable_ui(True)

    def init_menu(self):
        menu_bar = wx.MenuBar()
        file_menu = wx.Menu()
        help_menu = wx.Menu()

        quit_menu_item = wx.MenuItem(file_menu, APP_QUIT,
                                          '&Quit\tCtrl+Q')
        about_menu_item = wx.MenuItem(help_menu, APP_ABOUT,
                                           '&About ' + TITLE)

        file_menu.Append(quit_menu_item)
        help_menu.Append(about_menu_item)

        menu_bar.Append(file_menu, '&File')
        menu_bar.Append(help_menu, '&About')

        self.Bind(wx.EVT_MENU, self.on_quit, id=APP_QUIT)
        self.Bind(wx.EVT_MENU, self.on_help_btn, id=APP_ABOUT)
        self.SetMenuBar(menu_bar)

    def on_quit(self, e):
        self.Destroy()

    def enable_ui(self, enable):
        """Turn UI elements on and off."""
        self.xlsform_btn.Enable(enable)
        self.settings_btn.Enable(enable)
        self.output_btn.Enable(enable)
        self.help_btn.Enable(enable)
        self.run_btn.Enable(enable)

    @staticmethod
    def shorten_string(string, max_length):
        if len(string) >= max_length:
            return '...' + string[len(string) - max_length:len(string)]
        else:
            return string


def create_app():
    app = wx.App()
    app.SetAppName(TITLE)
    ex = MainWindow(None, size=(MAIN_WINDOW_WIDTH, MAIN_WINDOW_HEIGHT),
                    title=TITLE)
    ex.Show()
    app.MainLoop()
